from . import fields as fields
from . import validators as validators
from .app import APIFlask as APIFlask
from .blueprint import APIBlueprint as APIBlueprint
from .exceptions import abort as abort
from .exceptions import HTTPError as HTTPError
from .helpers import get_reason_phrase as get_reason_phrase
from .helpers import pagination_builder as pagination_builder
from .schemas import EmptySchema as EmptySchema
from .schemas import FileSchema as FileSchema
from .schemas import PaginationSchema as PaginationSchema
from .schemas import Schema as Schema
from .security import APIKeyCookieAuth as APIKeyCookieAuth
from .security import APIKeyHeaderAuth as APIKeyHeaderAuth
from .security import APIKeyQueryAuth as APIKeyQueryAuth
from .security import HTTPBasicAuth as HTTPBasicAuth
from .security import HTTPTokenAuth as HTTPTokenAuth
from .security import MultiAuth as MultiAuth



from apiflask import APIFlask, Schema, abort
from apiflask.fields import Integer, String
from apiflask.validators import Length, OneOf

app = APIFlask(__name__)
app.config['SYNC_LOCAL_SPEC'] = True
app.config['LOCAL_SPEC_PATH'] = 'openapi.json'
app.config['LOCAL_SPEC_JSON_INDENT'] = 0

pets = [{'id': 0, 'name': 'Kitty', 'category': 'cat'}, {'id': 1, 'name': 'Coco', 'category': 'dog'}]


class PetIn(Schema):
    name = String(required=True, validate=Length(0, 10))
    category = String(required=True, validate=OneOf(['dog', 'cat']))


class PetOut(Schema):
    id = Integer()
    name = String()
    category = String()


@app.get('/')
def say_hello():
    # returning a dict or list equals to use jsonify()
    return {'message': 'Hello!'}


@app.get('/pets/<int:pet_id>')
@app.output(PetOut)
def get_pet(pet_id):
    if pet_id > len(pets) - 1:
        abort(404)
    # you can also return an ORM/ODM model class instance directly
    # APIFlask will serialize the object into JSON format
    return pets[pet_id]


@app.patch('/pets/<int:pet_id>')
@app.input(PetIn(partial=True))  # -> json_data
@app.output(PetOut)
def update_pet(pet_id, json_data):
    # the validated and parsed input data will
    # be injected into the view function as a dict
    if pet_id > len(pets) - 1:
        abort(404)
    for attr, value in json_data.items():
        pets[pet_id][attr] = value
    return pets[pet_id]


if __name__ == '__main__':
    app.run()



import io

import openapi_spec_validator as osv
import pytest
from flask.views import MethodView
from packaging.version import Version
from werkzeug.datastructures import FileStorage

from .conftest import APISPEC_VERSION
from .schemas import Bar
from .schemas import EnumPathParameter
from .schemas import Files
from .schemas import Foo
from .schemas import Form
from .schemas import FormAndFiles
from apiflask import Schema
from apiflask.fields import String
from apiflask.validators import Length
from apiflask.validators import OneOf


def test_input(app, client):
    @app.route('/foo', methods=['POST'])
    @app.input(Foo)
    def foo(json_data):
        return json_data

    @app.route('/bar')
    class Bar(MethodView):
        @app.input(Foo)
        def post(self, json_data):
            return json_data

    for rule in ['/foo', '/bar']:
        rv = client.post(rule)
        assert rv.status_code == 422
        assert rv.json == {
            'detail': {'json': {'name': ['Missing data for required field.']}},
            'message': 'Validation error',
        }

        rv = client.post(rule, json={'id': 1})
        assert rv.status_code == 422
        assert rv.json == {
            'detail': {'json': {'name': ['Missing data for required field.']}},
            'message': 'Validation error',
        }

        rv = client.post(rule, json={'id': 1, 'name': 'bar'})
        assert rv.status_code == 200
        assert rv.json == {'id': 1, 'name': 'bar'}

        rv = client.post(rule, json={'name': 'bar'})
        assert rv.status_code == 200
        assert rv.json == {'name': 'bar'}

        rv = client.get('/openapi.json')
        assert rv.status_code == 200
        osv.validate(rv.json)
        assert (
            rv.json['paths'][rule]['post']['requestBody']['content']['application/json']['schema'][
                '$ref'
            ]
            == '#/components/schemas/Foo'
        )


def test_input_with_query_location(app, client):
    @app.route('/foo', methods=['POST'])
    @app.input(Foo, location='query', arg_name='foo')
    @app.input(Bar, location='query', arg_name='bar')
    def foo(foo, bar):
        return {'name': foo['name'], 'name2': bar['name2']}

    rv = client.post('/foo')
    assert rv.status_code == 422
    assert rv.json == {
        'detail': {'query': {'name': ['Missing data for required field.']}},
        'message': 'Validation error',
    }

    rv = client.post('/foo?id=1&name=bar')
    assert rv.status_code == 422
    assert rv.json == {
        'detail': {'query': {'name2': ['Missing data for required field.']}},
        'message': 'Validation error',
    }

    rv = client.post('/foo?id=1&name=bar&id2=2&name2=baz')
    assert rv.status_code == 200
    assert rv.json == {'name': 'bar', 'name2': 'baz'}

    rv = client.post('/foo?name=bar&name2=baz')
    assert rv.status_code == 200
    assert rv.json == {'name': 'bar', 'name2': 'baz'}


def test_input_with_form_location(app, client):
    @app.post('/')
    @app.input(Form, location='form')
    def index(form_data):
        return form_data

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert (
        'application/x-www-form-urlencoded'
        in rv.json['paths']['/']['post']['requestBody']['content']
    )
    assert (
        rv.json['paths']['/']['post']['requestBody']['content'][
            'application/x-www-form-urlencoded'
        ]['schema']['$ref']
        == '#/components/schemas/Form'
    )
    assert 'Form' in rv.json['components']['schemas']

    rv = client.post('/', data={'name': 'foo'})
    assert rv.status_code == 200
    assert rv.json == {'name': 'foo'}


def test_input_with_files_location(app, client):
    @app.post('/')
    @app.input(Files, location='files')
    def index(files_data):
        data = {}
        if 'image' in files_data and isinstance(files_data['image'], FileStorage):
            data['image'] = True
        return data

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'multipart/form-data' in rv.json['paths']['/']['post']['requestBody']['content']
    assert (
        rv.json['paths']['/']['post']['requestBody']['content']['multipart/form-data']['schema'][
            '$ref'
        ]
        == '#/components/schemas/Files'
    )
    assert 'image' in rv.json['components']['schemas']['Files']['properties']
    assert rv.json['components']['schemas']['Files']['properties']['image']['type'] == 'string'
    assert rv.json['components']['schemas']['Files']['properties']['image']['format'] == 'binary'

    rv = client.post(
        '/',
        data={
            'image': (io.BytesIO(b'test'), 'test.jpg'),
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 200
    assert rv.json == {'image': True}


def test_input_with_form_and_files_location(app, client):
    @app.post('/')
    @app.input(FormAndFiles, location='form_and_files')
    def index(form_and_files_data):
        data = {}
        if 'name' in form_and_files_data:
            data['name'] = True
        if 'image' in form_and_files_data and isinstance(form_and_files_data['image'], FileStorage):
            data['image'] = True
        return data

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'multipart/form-data' in rv.json['paths']['/']['post']['requestBody']['content']
    assert (
        rv.json['paths']['/']['post']['requestBody']['content']['multipart/form-data']['schema'][
            '$ref'
        ]
        == '#/components/schemas/FormAndFiles'
    )
    assert 'name' in rv.json['components']['schemas']['FormAndFiles']['properties']
    assert 'image' in rv.json['components']['schemas']['FormAndFiles']['properties']
    assert (
        rv.json['components']['schemas']['FormAndFiles']['properties']['image']['type'] == 'string'
    )
    assert (
        rv.json['components']['schemas']['FormAndFiles']['properties']['image']['format']
        == 'binary'
    )

    rv = client.post(
        '/',
        data={'name': 'foo', 'image': (io.BytesIO(b'test'), 'test.jpg')},
        content_type='multipart/form-data',
    )
    assert rv.status_code == 200
    assert rv.json == {'name': True, 'image': True}


def test_input_with_json_or_form_location(app, client):
    @app.post('/')
    @app.input(Form, location='json_or_form')
    def index(json_or_form_data):
        return json_or_form_data

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert (
        'application/x-www-form-urlencoded'
        in rv.json['paths']['/']['post']['requestBody']['content']
    )
    assert 'application/json' in rv.json['paths']['/']['post']['requestBody']['content']
    assert (
        rv.json['paths']['/']['post']['requestBody']['content']['application/json']['schema'][
            '$ref'
        ]
        == '#/components/schemas/Form'
    )
    assert (
        rv.json['paths']['/']['post']['requestBody']['content'][
            'application/x-www-form-urlencoded'
        ]['schema']['$ref']
        == '#/components/schemas/Form'
    )
    assert 'Form' in rv.json['components']['schemas']

    rv = client.post('/', data={'name': 'foo'})
    assert rv.status_code == 200
    assert rv.json == {'name': 'foo'}

    rv = client.post('/', json={'name': 'foo'})
    assert rv.status_code == 200
    assert rv.json == {'name': 'foo'}


def test_input_with_path_location(app, client):
    @app.get('/<image_type>')
    @app.input(EnumPathParameter, location='path')
    def index(image_type, path_data):
        return {'image_type': image_type}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/{image_type}' in rv.json['paths']
    assert len(rv.json['paths']['/{image_type}']['get']['parameters']) == 1
    assert rv.json['paths']['/{image_type}']['get']['parameters'][0]['in'] == 'path'
    assert rv.json['paths']['/{image_type}']['get']['parameters'][0]['name'] == 'image_type'
    assert rv.json['paths']['/{image_type}']['get']['parameters'][0]['schema'] == {
        'type': 'string',
        'enum': ['jpg', 'png', 'tiff', 'webp'],
    }

    rv = client.get('/png')
    assert rv.status_code == 200
    assert rv.json == {'image_type': 'png'}

    rv = client.get('/gif')
    assert rv.status_code == 422
    assert rv.json['message'] == 'Validation error'
    assert 'path' in rv.json['detail']
    assert 'image_type' in rv.json['detail']['path']
    assert rv.json['detail']['path']['image_type'] == ['Must be one of: jpg, png, tiff, webp.']


@pytest.mark.parametrize(
    'locations',
    [
        ['files', 'form'],
        ['files', 'json'],
        ['form', 'json'],
        ['form_and_files', 'json'],
        ['form_and_files', 'form'],
        ['form_and_files', 'files'],
        ['json_or_form', 'json'],
        ['json_or_form', 'files'],
        ['json_or_form', 'form'],
        ['json_or_form', 'form_and_files'],
    ],
)
def test_multiple_input_body_location(app, locations):
    arg_name_1 = f'{locations[0]}_data'  # noqa: F841
    arg_name_2 = f'{locations[1]}_data'  # noqa: F841
    with pytest.raises(RuntimeError):

        @app.route('/foo')
        @app.input(Foo, location=locations[0])
        @app.input(Bar, location=locations[1])
        def foo(arg_name_1, arg_name_2):
            pass


def test_input_with_dict_schema(app, client):
    dict_schema = {'name': String(required=True)}

    @app.get('/foo')
    @app.input(dict_schema, location='query')
    def foo(query_data):
        return query_data

    @app.post('/bar')
    @app.input(dict_schema, schema_name='MyName')
    def bar(json_data):
        return json_data

    @app.post('/baz')
    @app.input(dict_schema)
    def baz(json_data):
        return json_data

    @app.post('/spam')
    @app.input(dict_schema)
    def spam(json_data):
        return json_data

    rv = client.get('/foo')
    assert rv.status_code == 422
    assert rv.json == {
        'detail': {'query': {'name': ['Missing data for required field.']}},
        'message': 'Validation error',
    }

    rv = client.get('/foo?name=grey')
    assert rv.status_code == 200
    assert rv.json == {'name': 'grey'}

    rv = client.post('/bar')
    assert rv.status_code == 422
    assert rv.json == {
        'detail': {'json': {'name': ['Missing data for required field.']}},
        'message': 'Validation error',
    }

    rv = client.post('/bar', json={'name': 'grey'})
    assert rv.status_code == 200
    assert rv.json == {'name': 'grey'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['parameters'][0] == {
        'in': 'query',
        'name': 'name',
        'required': True,
        'schema': {'type': 'string'},
    }
    # TODO check the excess item "'x-scope': ['']" in schema object
    # https://github.com/p1c2u/openapi-spec-validator/issues/53
    assert (
        rv.json['paths']['/bar']['post']['requestBody']['content']['application/json']['schema'][
            '$ref'
        ]
        == '#/components/schemas/MyName'
    )

    # Check schema fields individually to avoid ordering issues
    schema = rv.json['components']['schemas']['MyName']
    assert schema['type'] == 'object'
    assert schema['properties'] == {'name': {'type': 'string'}}
    assert schema['required'] == ['name']
    # In apispec >= 6.8.3, additionalProperties should be False (or omitted, treated as False)
    if APISPEC_VERSION >= Version('6.8.3'):
        assert schema.get('additionalProperties', False) is False
    else:
        assert 'additionalProperties' not in schema
    # default schema name is "Generated"
    assert (
        rv.json['paths']['/baz']['post']['requestBody']['content']['application/json']['schema'][
            '$ref'
        ]
        == '#/components/schemas/GeneratedSchema'
    )
    assert (
        rv.json['paths']['/spam']['post']['requestBody']['content']['application/json']['schema'][
            '$ref'
        ]
        == '#/components/schemas/GeneratedSchema1'
    )


def test_input_body_example(app, client):
    example = {'name': 'foo', 'id': 2}
    examples = {
        'example foo': {'summary': 'an example of foo', 'value': {'name': 'foo', 'id': 1}},
        'example bar': {'summary': 'an example of bar', 'value': {'name': 'bar', 'id': 2}},
    }

    @app.post('/foo')
    @app.input(Foo, example=example)
    def foo():
        pass

    @app.post('/bar')
    @app.input(Foo, examples=examples)
    def bar():
        pass

    @app.route('/baz')
    class Baz(MethodView):
        @app.input(Foo, example=example)
        def get(self):
            pass

        @app.input(Foo, examples=examples)
        def post(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert (
        rv.json['paths']['/foo']['post']['requestBody']['content']['application/json']['example']
        == example
    )
    assert (
        rv.json['paths']['/bar']['post']['requestBody']['content']['application/json']['examples']
        == examples
    )

    assert (
        rv.json['paths']['/baz']['get']['requestBody']['content']['application/json']['example']
        == example
    )
    assert (
        rv.json['paths']['/baz']['post']['requestBody']['content']['application/json']['examples']
        == examples
    )


def test_skip_validation(app, client):
    incorrect_json = {'name': 'Kitty', 'category': 'unknown'}

    class PetIn(Schema):
        name = String(required=True, validate=Length(0, 10))
        category = String(required=True, validate=OneOf(['dog', 'cat']))

    @app.patch('/pets_without_validation/<int:pet_id>')
    @app.input(PetIn, validation=False)
    def pets_without_validation(pet_id, json_data):
        return {'pet_id': pet_id, 'json_data': json_data}

    no_validated_rv = client.patch('/pets_without_validation/1', json=incorrect_json)
    assert no_validated_rv.status_code == 200
    assert no_validated_rv.json['json_data']['name'] == 'Kitty'
    assert no_validated_rv.json['json_data']['category'] == 'unknown'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    assert (
        rv.json['paths']['/pets_without_validation/{pet_id}']['patch']['requestBody']['content'][
            'application/json'
        ]['schema']['$ref']
        == '#/components/schemas/PetIn'
    )
    assert 'PetIn' in rv.json['components']['schemas']


@pytest.mark.parametrize('validation', [True, False])
@pytest.mark.parametrize('payload', [[], [{'bar': 'baz'}], [{'qux': 'baz'}]])
def test_skip_validation_list_input(app, client, validation, payload):
    class FooIn(Schema):
        bar = String(required=True)

    @app.put('/foo/bulk')
    @app.input(FooIn(many=True), validation=validation)
    def bulk_put_foo(json_data):
        return json_data

    rv = client.put('/foo/bulk', json=payload)
    if validation and payload and 'bar' not in payload[0]:
        assert rv.status_code == 422
    else:
        assert rv.status_code == 200
        assert rv.json == payload


@pytest.mark.parametrize('validation', [True, False])
@pytest.mark.parametrize('payload', [{}, {'bar': 'qux'}])
def test_skip_validation_arg_name(app, client, validation, payload):
    class FooIn(Schema):
        bar = String(required=True)

    @app.post('/foo')
    @app.input(FooIn, arg_name='baz', validation=validation)
    def post_foo(baz):
        return baz

    rv = client.post('/foo', json=payload)
    if validation and not payload:
        assert rv.status_code == 422
    else:
        assert rv.status_code == 200
        assert rv.json == payload



from __future__ import annotations

import typing as t

from .fields import DelimitedList
from .schema_adapters import registry

if t.TYPE_CHECKING:
    from apispec import APISpec
    from apispec.ext.marshmallow import MarshmallowPlugin


def get_unique_schema_name(spec: APISpec, base_name: str) -> str:
    """Generate a unique schema name by appending a counter.

    Arguments:
        spec: The APISpec object
        base_name: The base schema name

    Returns:
        A unique schema name that doesn't exist in spec.components.schemas
    """
    counter = 1
    schema_name = base_name
    while schema_name in spec.components.schemas:
        schema_name = f'{base_name}{counter}'
        counter += 1
    return schema_name


def extract_pydantic_defs(schema_dict: dict[str, t.Any], parent_name: str) -> dict[str, t.Any]:
    """Extract $defs from Pydantic schema and return them with composed keys.

    Arguments:
        schema_dict: Schema dictionary from Pydantic model_json_schema()
        parent_name: Parent schema name for composing nested schema names

    Returns:
        Dictionary of extracted definitions with composed keys like "ParentName.ChildName"
    """
    definitions = {}

    if '$defs' in schema_dict:
        for def_name, def_schema in schema_dict['$defs'].items():
            composed_key = f'{parent_name}.{def_name}'
            definitions[composed_key] = def_schema

        # Remove $defs from the original schema
        del schema_dict['$defs']

    return definitions


class OpenAPIHelper:
    """Helper class for generating OpenAPI schemas from different schema types.

    *Version added: 3.0.0*
    """

    def __init__(self) -> None:
        self._marshmallow_plugin: MarshmallowPlugin | None = None
        self._marshmallow_plugin_initialized = False

    def get_marshmallow_plugin(self) -> MarshmallowPlugin | None:
        """Get or create marshmallow plugin for OpenAPI schema generation.

        Returns a MarshmallowPlugin with initialized converter, ready to use.
        Returns None if marshmallow is not installed.
        """
        if not self._marshmallow_plugin_initialized:
            try:
                from apispec import APISpec
                from apispec.ext.marshmallow import MarshmallowPlugin

                self._marshmallow_plugin = MarshmallowPlugin()
                # Initialize the plugin's converter by adding it to an APISpec
                APISpec(
                    title='_temp',
                    version='1.0.0',
                    openapi_version='3.0.3',
                    plugins=[self._marshmallow_plugin],
                )
                self._marshmallow_plugin.converter.add_parameter_attribute_function(  # type: ignore
                    self.delimited_list2param
                )

            except ImportError:
                self._marshmallow_plugin = None
            finally:
                self._marshmallow_plugin_initialized = True

        return self._marshmallow_plugin

    def schema_to_spec(self, schema: t.Any) -> dict[str, t.Any]:
        """Convert a schema to OpenAPI specification.

        Arguments:
            schema: Schema object (marshmallow, Pydantic, etc.)

        Returns:
            OpenAPI schema dict
        """
        try:
            adapter = registry.create_adapter(schema)
            return adapter.get_openapi_schema()
        except Exception:
            # Fallback for unknown schema types
            return {'type': 'object'}

    def schema_to_json_schema(self, schema: t.Any) -> dict[str, t.Any]:
        """Convert a schema to full JSON schema with properties.

        This is different from schema_to_spec in that it returns the complete
        JSON schema definition including all properties, rather than potentially
        returning a reference. Used for base response schemas and other cases
        where the full schema definition is needed.

        Arguments:
            schema: Schema object (marshmallow, Pydantic, etc.)

        Returns:
            Full JSON schema dict with properties
        """
        try:
            adapter = registry.create_adapter(schema)

            # For marshmallow schemas, use schema2jsonschema
            if adapter.schema_type == 'marshmallow':
                plugin = self.get_marshmallow_plugin()
                if plugin is not None:
                    return plugin.converter.schema2jsonschema(adapter.schema)  # type: ignore[union-attr, no-any-return]

            # For other schema types, fall back to get_openapi_schema
            return adapter.get_openapi_schema()
        except Exception:
            # Fallback for unknown schema types
            return {'type': 'object'}

    def schema_to_parameters(
        self, schema: t.Any, location: str = 'query'
    ) -> list[dict[str, t.Any]]:
        """Convert schema to OpenAPI parameters.

        Arguments:
            schema: Schema object
            location: Parameter location ('query', 'header', etc.)

        Returns:
            List of OpenAPI parameter definitions
        """
        try:
            adapter = registry.create_adapter(schema)

            # Map location to OpenAPI 'in' field
            openapi_location = location
            if location == 'headers':
                openapi_location = 'header'
            elif location == 'view_args':
                openapi_location = 'path'
            elif location == 'querystring':
                openapi_location = 'query'
            elif location == 'cookies':
                openapi_location = 'cookie'

            # For marshmallow schemas, extract parameters directly from fields
            if adapter.schema_type == 'marshmallow':
                return self._extract_marshmallow_parameters(
                    adapter.schema, location=openapi_location
                )

            # For other schema types, generate basic parameters
            schema_spec = adapter.get_openapi_schema()
            parameters = []

            if 'properties' in schema_spec:
                required = schema_spec.get('required', [])
                for name, prop_spec in schema_spec['properties'].items():
                    param = {
                        'name': name,
                        'in': openapi_location,
                        'required': name in required,
                        'schema': prop_spec,
                    }
                    parameters.append(param)

            return parameters
        except Exception:
            return []

    def delimited_list2param(self, field, **kwargs) -> dict:  # type: ignore[no-untyped-def]
        """Set correct OpenAPI parameter attributes for DelimitedList fields."""
        ret: dict = {}
        if isinstance(field, DelimitedList):
            ret['explode'] = False
            ret['style'] = 'form'
        return ret

    def _extract_marshmallow_parameters(
        self, schema: t.Any, location: str
    ) -> list[dict[str, t.Any]]:
        """Extract parameters from marshmallow schema fields using apispec converter."""
        plugin = self.get_marshmallow_plugin()
        if plugin is None:
            return []

        return plugin.converter.schema2parameters(schema, location=location)  # type: ignore[union-attr, no-any-return]

    def get_schema_name(self, schema: t.Any) -> str:
        """Get the name for a schema.

        Arguments:
            schema: Schema object

        Returns:
            Schema name string
        """
        try:
            adapter = registry.create_adapter(schema)
            return adapter.get_schema_name()
        except Exception:
            return 'Schema'


# Global helper instance
openapi_helper = OpenAPIHelper()



from apiflask import APIBlueprint


def test_app_decorators(app):
    assert hasattr(app, 'auth_required')
    assert hasattr(app, 'input')
    assert hasattr(app, 'output')
    assert hasattr(app, 'doc')


def test_bp_decorators(app):
    bp = APIBlueprint('test', __name__)
    assert hasattr(bp, 'auth_required')
    assert hasattr(bp, 'input')
    assert hasattr(bp, 'output')
    assert hasattr(bp, 'doc')



from __future__ import annotations

import typing as t

from marshmallow import Schema as BaseSchema
from marshmallow.fields import Integer
from marshmallow.fields import URL


# schema for the detail object of validation error response
validation_error_detail_schema: dict[str, t.Any] = {
    'type': 'object',
    'properties': {
        '<location>': {
            'type': 'object',
            'properties': {'<field_name>': {'type': 'array', 'items': {'type': 'string'}}},
        }
    },
}


# schema for validation error response
validation_error_schema: dict[str, t.Any] = {
    'properties': {
        'detail': validation_error_detail_schema,
        'message': {'type': 'string'},
    },
    'type': 'object',
}


# schema for generic error response
http_error_schema: dict[str, t.Any] = {
    'properties': {
        'detail': {'type': 'object'},
        'message': {'type': 'string'},
    },
    'type': 'object',
}


class Schema(BaseSchema):
    """A base schema for all schemas. Equivalent to `marshmallow.Schema`.

    *Version Added: 1.2.0*
    """

    pass


class EmptySchema(Schema):
    """An empty schema used to generate empty response/schema.

    *Version changed: 3.0.0*

    - Removed from docs and should only be used internally. Use `{}` instead.
    """

    pass


class PaginationSchema(Schema):
    """A schema for common pagination information."""

    page = Integer()
    per_page = Integer()
    pages = Integer()
    total = Integer()
    current = URL()
    next = URL()
    prev = URL()
    first = URL()
    last = URL()


class FileSchema(Schema):
    """A schema for file response.

    This is used to represent a file response in OpenAPI spec. If you want to
    embed a file as base64 string in the JSON body, you can use the
    `apiflask.fields.File` field instead.

    Example:

    ```python
    from apiflask.schemas import FileSchema
    from flask import send_from_directory

    @app.get('/images/<filename>')
    @app.output(
        FileSchema(type='string', format='binary'),
        content_type='image/png',
        description='An image file'
    )
    @app.doc(summary="Returns the image file")
    def get_image(filename):
        return send_from_directory(app.config['IMAGE_FOLDER'], filename)
    ```

    The output OpenAPI spec will be:

    ```yaml
    paths:
    /images/{filename}:
      get:
        summary: Returns the image file
        responses:
          '200':
            description: An image file
            content:
              image/png:
                schema:
                  type: string
                  format: binary
    ```

    *Version Added: 2.0.0*
    """

    def __init__(self, *, type: str = 'string', format: str = 'binary') -> None:
        """
        Arguments:
            type: The type of the file. Defaults to `string`.
            format: The format of the file, one of `binary` and `base64`. Defaults to `binary`.
        """
        self.type = type
        self.format = format

    def __repr__(self) -> str:
        return f'schema: \n  type: {self.type}\n  format: {self.format}'



import openapi_spec_validator as osv
import pytest

from .schemas import Foo
from apiflask import Schema
from apiflask.fields import Field
from apiflask.fields import Integer
from apiflask.fields import String


class BaseResponse(Schema):
    message = String()
    status_code = Integer()
    data = Field()


class BadBaseResponse(Schema):
    message = String()
    status_code = Integer()
    some_data = Field()


base_response_schema_dict = {
    'properties': {
        'data': {'type': 'object'},
        'message': {'type': 'string'},
        'status_code': {'type': 'integer'},
    },
    'type': 'object',
}

bad_base_response_schema_dict = {
    'properties': {
        'some_data': {'type': 'object'},
        'message': {'type': 'string'},
        'status_code': {'type': 'integer'},
    },
    'type': 'object',
}


def test_output_base_response(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    @app.get('/')
    @app.output(Foo)
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'data': data}

    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json
    assert 'data' in rv.json
    assert rv.json['message'] == 'Success.'
    assert rv.json['status_code'] == 200
    assert rv.json['data']['id'] == 123
    assert rv.json['data']['name'] == 'test'


@pytest.mark.parametrize(
    'base_schema',
    [
        BaseResponse,
        base_response_schema_dict,
        BadBaseResponse,
        bad_base_response_schema_dict,
        '',
        None,
    ],
)
def test_base_response_spec(app, client, base_schema):
    app.config['BASE_RESPONSE_SCHEMA'] = base_schema

    @app.get('/')
    @app.output(Foo)
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'data': data}

    if base_schema == '':
        with pytest.raises(TypeError) as e:
            app.spec
        assert 'marshmallow' in str(e.value)
    elif base_schema in [BadBaseResponse, bad_base_response_schema_dict]:
        with pytest.raises(RuntimeError) as e:
            app.spec
        assert 'data key' in str(e.value)
    else:
        rv = client.get('/openapi.json')
        assert rv.status_code == 200
        osv.validate(rv.json)
        schema = rv.json['paths']['/']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]
        schema_ref = '#/components/schemas/Foo'
        # TODO the output schema ref contains unused `'x-scope': ['']` field
        # it seems related to openapi-spec-validator:
        # https://github.com/p1c2u/openapi-spec-validator/issues/53
        if base_schema in [BaseResponse, base_response_schema_dict]:
            properties = schema['properties']
            assert properties['data']['$ref'] == schema_ref
            assert properties['status_code'] == {'type': 'integer'}
            assert properties['message'] == {'type': 'string'}
        elif base_schema is None:
            assert schema['$ref'] == schema_ref


def test_base_response_data_key(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse
    app.config['BASE_RESPONSE_DATA_KEY '] = 'data'

    @app.get('/')
    @app.output(Foo)
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'info': data}

    with app.test_request_context('/'):
        with pytest.raises(RuntimeError, match=r'The data key.*is not found in the returned dict.'):
            foo()


def test_base_response_204(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    @app.get('/')
    @app.output({}, status_code=204)
    def foo():
        return ''

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'content' not in rv.json['paths']['/']['get']['responses']['204']


def test_base_response_many(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    @app.get('/')
    @app.output(Foo(many=True))
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'data': data}

    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json
    assert 'data' in rv.json
    assert not isinstance(rv.json, list)
    assert isinstance(rv.json['data'], list)


def test_bare_view_base_response_spec(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    @app.get('/')
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'data': data}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    schema = rv.json['paths']['/']['get']['responses']['200']['content']['application/json'][
        'schema'
    ]
    assert schema['properties']['status_code'] == {'type': 'integer'}
    assert schema['properties']['message'] == {'type': 'string'}


def test_input_with_base_response_spec(app, client):
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    @app.get('/')
    @app.input(Foo)
    def foo():
        data = {'id': '123', 'name': 'test'}
        return {'message': 'Success.', 'status_code': '200', 'data': data}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    schema = rv.json['paths']['/']['get']['responses']['200']['content']['application/json'][
        'schema'
    ]
    assert schema['properties']['status_code'] == {'type': 'integer'}
    assert schema['properties']['message'] == {'type': 'string'}



import openapi_spec_validator as osv
import pytest
from flask.views import MethodView
from packaging.version import Version

from .conftest import APISPEC_VERSION
from .schemas import CustomHTTPError
from .schemas import Foo


def test_doc_summary_and_description(app, client):
    @app.route('/foo')
    @app.doc(summary='summary from doc decorator')
    def foo():
        pass

    @app.route('/bar')
    @app.doc(summary='summary for bar', description='some description for bar')
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['summary'] == 'summary from doc decorator'
    assert 'description' not in rv.json['paths']['/foo']['get']
    assert rv.json['paths']['/bar']['get']['summary'] == 'summary for bar'
    assert rv.json['paths']['/bar']['get']['description'] == 'some description for bar'


def test_doc_summary_and_description_with_methodview(app, client):
    @app.route('/baz')
    class Baz(MethodView):
        @app.doc(summary='summary from doc decorator')
        def get(self):
            pass

        @app.doc(summary='summary for baz', description='some description for baz')
        def post(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/baz']['get']['summary'] == 'summary from doc decorator'
    assert 'description' not in rv.json['paths']['/baz']['get']
    assert rv.json['paths']['/baz']['post']['summary'] == 'summary for baz'
    assert rv.json['paths']['/baz']['post']['description'] == 'some description for baz'


def test_doc_tags(app, client):
    app.tags = ['foo', 'bar']

    @app.route('/foo')
    @app.doc(tags=['foo'])
    def foo():
        pass

    @app.route('/bar')
    @app.doc(tags=['foo', 'bar'])
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['tags'] == ['foo']
    assert rv.json['paths']['/bar']['get']['tags'] == ['foo', 'bar']


def test_doc_tags_with_methodview(app, client):
    @app.route('/baz')
    class Baz(MethodView):
        @app.doc(tags=['foo'])
        def get(self):
            pass

        @app.doc(tags=['foo', 'bar'])
        def post(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/baz']['get']['tags'] == ['foo']
    assert rv.json['paths']['/baz']['post']['tags'] == ['foo', 'bar']


def test_doc_hide(app, client):
    @app.route('/foo')
    @app.doc(hide=True)
    def foo():
        pass

    @app.get('/baz')
    def get_baz():
        pass

    @app.post('/baz')
    @app.doc(hide=True)
    def post_baz():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/foo' not in rv.json['paths']
    assert '/baz' in rv.json['paths']
    assert 'get' in rv.json['paths']['/baz']
    assert 'post' not in rv.json['paths']['/baz']


def test_doc_hide_with_methodview(app, client):
    @app.route('/bar')
    class Bar(MethodView):
        def get(self):
            pass

        @app.doc(hide=True)
        def post(self):
            pass

    @app.route('/secret')
    class Secret(MethodView):
        @app.doc(hide=True)
        def get(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/bar' in rv.json['paths']
    assert 'get' in rv.json['paths']['/bar']
    assert 'post' not in rv.json['paths']['/bar']
    assert '/secret' in rv.json['paths']


def test_doc_deprecated(app, client):
    @app.route('/foo')
    @app.doc(deprecated=True)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['deprecated']


def test_doc_deprecated_with_methodview(app, client):
    @app.route('/foo')
    class FooAPI(MethodView):
        @app.doc(deprecated=True)
        def get(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['deprecated']


def test_doc_responses(app, client):
    @app.route('/foo')
    @app.input(Foo)
    @app.output(Foo)
    @app.doc(responses={200: 'success', 400: 'bad', 404: 'not found', 500: 'server error'})
    def foo():
        pass

    @app.route('/bar')
    @app.input(Foo)
    @app.output(Foo)
    @app.doc(responses=[200, 400, 404, 500])
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '200' in rv.json['paths']['/foo']['get']['responses']
    assert '400' in rv.json['paths']['/foo']['get']['responses']
    # overwrite existing error descriptions
    assert rv.json['paths']['/foo']['get']['responses']['200']['description'] == 'success'
    assert rv.json['paths']['/foo']['get']['responses']['400']['description'] == 'bad'
    assert '404' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/foo']['get']['responses']['404']['description'] == 'not found'
    assert '500' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/foo']['get']['responses']['500']['description'] == 'server error'

    assert '200' in rv.json['paths']['/bar']['get']['responses']
    assert '400' in rv.json['paths']['/bar']['get']['responses']
    assert (
        rv.json['paths']['/bar']['get']['responses']['200']['description'] == 'Successful response'
    )
    assert rv.json['paths']['/bar']['get']['responses']['400']['description'] == 'Bad Request'
    assert '404' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/bar']['get']['responses']['404']['description'] == 'Not Found'
    assert '500' in rv.json['paths']['/bar']['get']['responses']
    assert (
        rv.json['paths']['/bar']['get']['responses']['500']['description']
        == 'Internal Server Error'
    )


def test_doc_responses_custom_spec(app, client):
    response_spec = {
        'description': 'Success',
        'content': {
            'application/json': {
                'schema': {
                    'type': 'object',
                    'properties': {
                        'data': {
                            'type': 'object',
                            'properties': {
                                'id': {'type': 'integer'},
                                'name': {'type': 'string'},
                            },
                        }
                    },
                }
            }
        },
    }

    @app.get('/foo')
    @app.input(Foo)
    @app.output(Foo)
    @app.doc(responses={200: response_spec})
    def foo():
        return {'message': 'Hello!'}

    @app.get('/bar')
    @app.doc(
        responses={
            200: {'description': 'Success', 'content': {'application/json': {'schema': Foo}}},
            400: {
                'description': 'Error',
                'content': {'application/json': {'schema': CustomHTTPError}},
            },
            404: {
                'description': 'Error',
                'content': {'application/json': {'schema': CustomHTTPError()}},
            },
        }
    )
    def say_hello():
        return {'message': 'Hello!'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '200' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/foo']['get']['responses']['200'] == response_spec

    assert '200' in rv.json['paths']['/bar']['get']['responses']
    assert '400' in rv.json['paths']['/bar']['get']['responses']
    assert '404' in rv.json['paths']['/bar']['get']['responses']
    assert rv.json['paths']['/bar']['get']['responses']['200']['description'] == 'Success'
    assert rv.json['paths']['/bar']['get']['responses']['400']['description'] == 'Error'
    assert rv.json['paths']['/bar']['get']['responses']['400']['content']['application/json'][
        'schema'
    ] == {'$ref': '#/components/schemas/CustomHTTPError'}

    # Check schema fields individually to avoid ordering issues
    schema = rv.json['components']['schemas']['CustomHTTPError']
    assert schema['type'] == 'object'
    assert schema['properties'] == {
        'status_code': {'type': 'string'},
        'message': {'type': 'string'},
        'custom': {'type': 'string'},
    }
    assert set(schema['required']) == {'custom', 'message', 'status_code'}
    # In apispec >= 6.8.3, additionalProperties should be False (or omitted, treated as False)
    if APISPEC_VERSION >= Version('6.8.3'):
        assert schema.get('additionalProperties', False) is False
    else:
        assert 'additionalProperties' not in schema


def test_doc_responses_additional_content_type(app, client):
    """Verify that it is possible to add additional media types for a response's status code."""
    description = 'something'

    @app.route('/foo')
    @app.input(Foo)
    @app.output(Foo)
    @app.doc(
        responses={
            200: {
                'content': {
                    'text/html': {},
                },
            },
        }
    )
    def foo():
        pass

    @app.route('/bar')
    @app.input(Foo)
    @app.output(Foo)
    @app.doc(
        responses={
            200: {
                'content': {
                    'text/html': {},
                },
                'description': description,
            },
        }
    )
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '200' in rv.json['paths']['/foo']['get']['responses']
    assert 'application/json' in rv.json['paths']['/foo']['get']['responses']['200']['content']
    assert 'text/html' in rv.json['paths']['/foo']['get']['responses']['200']['content']
    assert (
        rv.json['paths']['/foo']['get']['responses']['200']['description'] == 'Successful response'
    )  # noqa: E501
    assert '200' in rv.json['paths']['/bar']['get']['responses']
    assert 'application/json' in rv.json['paths']['/bar']['get']['responses']['200']['content']
    assert 'text/html' in rv.json['paths']['/bar']['get']['responses']['200']['content']
    assert rv.json['paths']['/bar']['get']['responses']['200']['description'] == description


def test_doc_responses_with_methodview(app, client):
    @app.route('/foo')
    class FooAPI(MethodView):
        @app.input(Foo)
        @app.output(Foo)
        @app.doc(responses={200: 'success', 400: 'bad', 404: 'not found', 500: 'server error'})
        def get(self):
            pass

    @app.route('/bar')
    class BarAPI(MethodView):
        @app.input(Foo)
        @app.output(Foo)
        @app.doc(responses=[200, 400, 404, 500])
        def get(self):
            pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '200' in rv.json['paths']['/foo']['get']['responses']
    assert '400' in rv.json['paths']['/foo']['get']['responses']
    # don't overwrite exist error description
    assert rv.json['paths']['/foo']['get']['responses']['200']['description'] == 'success'
    assert rv.json['paths']['/foo']['get']['responses']['400']['description'] == 'bad'
    assert '404' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/foo']['get']['responses']['404']['description'] == 'not found'
    assert '500' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/foo']['get']['responses']['500']['description'] == 'server error'

    assert '200' in rv.json['paths']['/bar']['get']['responses']
    assert '400' in rv.json['paths']['/bar']['get']['responses']
    assert (
        rv.json['paths']['/bar']['get']['responses']['200']['description'] == 'Successful response'
    )
    assert rv.json['paths']['/bar']['get']['responses']['400']['description'] == 'Bad Request'
    assert '404' in rv.json['paths']['/foo']['get']['responses']
    assert rv.json['paths']['/bar']['get']['responses']['404']['description'] == 'Not Found'
    assert '500' in rv.json['paths']['/bar']['get']['responses']
    assert (
        rv.json['paths']['/bar']['get']['responses']['500']['description']
        == 'Internal Server Error'
    )


def test_doc_operationid(app, client):
    @app.route('/foo')
    @app.doc(operation_id='getSomeFoo')
    def foo():
        pass

    @app.route('/bar')
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['operationId'] == 'getSomeFoo'
    assert 'operationId' not in rv.json['paths']['/bar']['get']


def test_doc_security(app, client):
    @app.route('/foo')
    @app.doc(security='ApiKeyAuth')
    def foo():
        pass

    @app.route('/bar')
    @app.doc(security=['BasicAuth', 'ApiKeyAuth'])
    def bar():
        pass

    @app.route('/baz')
    @app.doc(security=[{'OAuth2': ['read', 'write']}])
    def baz():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['security'] == [{'ApiKeyAuth': []}]
    assert rv.json['paths']['/bar']['get']['security'] == [{'BasicAuth': []}, {'ApiKeyAuth': []}]
    assert rv.json['paths']['/baz']['get']['security'] == [{'OAuth2': ['read', 'write']}]


def test_doc_security_invalid_value(app):
    @app.route('/foo')
    @app.doc(security={'BasicAuth': []})
    def foo():
        pass

    with pytest.raises(ValueError):
        app.spec



import openapi_spec_validator as osv
import pytest

from apiflask.schemas import EmptySchema
from apiflask.schemas import FileSchema


def test_file_schema(app, client):
    @app.get('/image')
    @app.output(
        FileSchema(type='string', format='binary'),
        content_type='image/png',
        description='An image file',
    )
    def get_image():
        return 'file'

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    content = rv.json['paths']['/image']['get']['responses']['200']['content']
    assert 'image/png' in content
    assert content['image/png']['schema'] == {'type': 'string', 'format': 'binary'}
    rv = client.get('/image')
    assert rv.status_code == 200


def test_file_schema_repr():
    f = FileSchema(type='string', format='binary')
    assert repr(f) == f'schema: \n  type: {f.type}\n  format: {f.format}'


@pytest.mark.parametrize('schema', [{}, EmptySchema])
def test_empty_schema(app, client, schema):
    @app.route('/foo')
    @app.output(schema, status_code=204)
    def empty_reposonse():
        return ''

    @app.route('/bar')
    @app.output(schema, content_type='image/png')
    def empty_schema():
        return ''

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'content' not in rv.json['paths']['/foo']['get']['responses']['204']

    assert 'content' in rv.json['paths']['/bar']['get']['responses']['200']
    assert (
        rv.json['paths']['/bar']['get']['responses']['200']['content']['image/png']['schema'] == {}
    )



import os
from importlib import metadata

import pytest
from packaging.version import Version

from apiflask import APIFlask

# Get apispec version once for all tests
APISPEC_VERSION = Version(metadata.version('apispec'))


@pytest.fixture
def app():
    app = APIFlask(__name__)
    return app


@pytest.fixture
def client(app):
    return app.test_client()


@pytest.fixture
def cli_runner(app):
    return app.test_cli_runner()


@pytest.fixture
def test_apps(monkeypatch):
    monkeypatch.syspath_prepend(
        os.path.abspath(os.path.join(os.path.dirname(__file__), 'test_apps'))
    )



import io
from importlib.metadata import version

import pytest
from packaging.version import parse
from werkzeug.datastructures import FileStorage

from .schemas import Files
from .schemas import FilesList
from apiflask import Schema
from apiflask.fields import Config


def test_file_field(app, client):
    @app.post('/')
    @app.input(Files, location='files')
    def index(files_data):
        data = {}
        if 'image' in files_data and isinstance(files_data['image'], FileStorage):
            data['image'] = True
        return data

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    assert 'image' in rv.json['components']['schemas']['Files']['properties']
    assert rv.json['components']['schemas']['Files']['properties']['image']['type'] == 'string'
    assert rv.json['components']['schemas']['Files']['properties']['image']['format'] == 'binary'

    rv = client.post(
        '/',
        data={
            'image': (io.BytesIO(b'test'), 'test.jpg'),
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 200
    assert rv.json == {'image': True}

    rv = client.post(
        '/',
        data={
            'image': 'test',
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 422
    assert rv.json['detail']['files']['image'] == ['Not a valid file.']


@pytest.mark.skipif(
    parse(version('flask-marshmallow')) < parse('1.2.1'),
    reason='Need flask_marshmallow 1.2.1 for processing empty file field',
)
def test_empty_file_field(app, client):
    @app.post('/')
    @app.input(Files, location='files')
    def index(files_data):
        data = {}
        if 'image' in files_data and isinstance(files_data['image'], FileStorage):
            data['image'] = True
        return data

    rv = client.post(
        '/',
        data={
            'image': '',
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 200
    assert rv.json == {}


def test_multiple_file_field(app, client):
    @app.post('/')
    @app.input(FilesList, location='files')
    def index(files_data):
        data = {'images': True}
        for f in files_data['images']:
            if not isinstance(f, FileStorage):
                data['images'] = False
        return data

    rv = client.post(
        '/',
        data={
            'images': [
                (io.BytesIO(b'test0'), 'test0.jpg'),
                (io.BytesIO(b'test1'), 'test1.jpg'),
                (io.BytesIO(b'test2'), 'test2.jpg'),
            ]
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 200
    assert rv.json == {'images': True}

    rv = client.post(
        '/',
        data={
            'images': [
                (io.BytesIO(b'test0'), 'test0.jpg'),
                (io.BytesIO(b'test1'), 'test1.jpg'),
                'test2',
            ]
        },
        content_type='multipart/form-data',
    )
    assert rv.status_code == 422
    assert rv.json['detail']['files']['images']['2'] == ['Not a valid file.']


def test_config_field(app, client):
    app.config['API_TITLE'] = 'Pet API'

    class GoodSchema(Schema):
        title = Config('API_TITLE')

    class BadSchema(Schema):
        description = Config('API_DESCRIPTION')

    @app.get('/good')
    @app.output(GoodSchema)
    def good():
        return {}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    rv = client.get('/good')
    assert rv.status_code == 200
    assert rv.json == {'title': 'Pet API'}

    with pytest.raises(ValueError, match=r'The key.*is not found in the app config.'):
        with app.app_context():
            bad = BadSchema()
            bad.dump({})



import typing as t
from flask import current_app
from apiflask import APIFlask, APIKeyHeaderAuth, abort
from authlib.jose import jwt, JoseError

app = APIFlask(__name__)
auth = APIKeyHeaderAuth()
app.config['SECRET_KEY'] = 'secret-key'


class User:
    def __init__(self, id: int, username: str):
        self.id = id
        self.username = username

    def get_apikey(self):
        header = {'alg': 'HS256'}
        payload = {'id': self.id}
        return jwt.encode(header, payload, current_app.config['SECRET_KEY']).decode()


users = [
    User(1, 'lorem'),
    User(2, 'ipsum'),
]


def get_user_by_id(id: int) -> t.Union[User, None]:
    return tuple(filter(lambda u: u.id == id, users))[0]


@auth.verify_token
def verify_apikey(apikey: str) -> t.Union[User, None]:
    try:
        data = jwt.decode(
            apikey.encode('ascii'),
            current_app.config['SECRET_KEY'],
        )
        id = data['id']
        user = get_user_by_id(id)
    except JoseError:
        return None
    except IndexError:
        return None
    return user


@app.post('/apikey/<int:id>')
def get_token(id: int):
    if get_user_by_id(id) is None:
        abort(404)
    return {'api_key': f'{get_user_by_id(id).get_apikey()}'}


@app.get('/name')
@app.auth_required(auth)
def get_username():
    return auth.current_user.username



import os
import sys
from importlib import metadata, reload

import pytest
from packaging.version import Version

full_examples = [
    'basic',
    'orm',
    'cbv',
    'openapi/basic',
    'blueprint_tags',
    'base_response/marshmallow',
    'base_response/pydantic',
    'pydantic',
    'dataclass',
]

examples = full_examples + ['pagination']
flask_sqlalchemy_version = Version(metadata.version('flask_sqlalchemy')).major


@pytest.fixture
def client(request):
    app_path = os.path.join(os.path.dirname(__file__), f'{request.param}')
    sys.path.insert(0, app_path)
    import app

    app = reload(app)
    _app = app.app
    _app.testing = True
    sys.path.remove(app_path)
    client = _app.test_client()
    client._example = request.param
    return client


@pytest.mark.parametrize('client', examples, indirect=True)
def test_say_hello(client):
    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json
    if client._example.startswith('base_response'):
        data = rv.json['data']
    else:
        data = rv.json
    assert data['message'] == 'Hello!'


@pytest.mark.skipif(flask_sqlalchemy_version < 3, reason='Need flask_sqlalchemy 3 for get_or_404')
@pytest.mark.parametrize('client', examples, indirect=True)
def test_get_pet(client):
    rv = client.get('/pets/1')
    assert rv.status_code == 200
    assert rv.json
    if client._example.startswith('base_response'):
        data = rv.json['data']
    else:
        data = rv.json
    if client._example == 'orm':
        assert data['name'] == 'Kitty'
        assert data['category'] == 'cat'
    elif client._example == 'pagination':
        assert data['name'] == 'Pet 1'
    else:
        assert data['name'] == 'Coco'
        assert data['category'] == 'dog'

    rv = client.get('/pets/13')
    if client._example != 'pagination':
        assert rv.status_code == 404
        assert rv.json


@pytest.mark.parametrize('client', full_examples, indirect=True)
def test_get_pets(client):
    rv = client.get('/pets')
    assert rv.status_code == 200
    assert rv.json
    if client._example.startswith('base_response'):
        data = rv.json['data']
    else:
        data = rv.json
    assert len(rv.json) == 3
    assert data[0]['name'] == 'Kitty'
    assert data[0]['category'] == 'cat'


@pytest.mark.parametrize('client', full_examples, indirect=True)
def test_create_pet(client):
    rv = client.post('/pets', json={'name': 'Grey', 'category': 'cat'})
    assert rv.status_code == 201
    assert rv.json
    if client._example.startswith('base_response'):
        data = rv.json['data']
    else:
        data = rv.json
    assert data['name'] == 'Grey'
    assert data['category'] == 'cat'


@pytest.mark.parametrize('client', full_examples, indirect=True)
@pytest.mark.parametrize(
    'data',
    [
        {'name': 'Grey', 'category': 'human'},
        {'name': 'Fyodor Mikhailovich Dostoevsky', 'category': 'cat'},
        {'category': 'cat'},
        {'name': 'Grey'},
    ],
)
def test_create_pet_with_bad_data(client, data):
    rv = client.post('/pets', json=data)
    assert rv.status_code == 422
    assert rv.json


@pytest.mark.skipif(flask_sqlalchemy_version < 3, reason='Need flask_sqlalchemy 3 for get_or_404')
@pytest.mark.parametrize('client', full_examples, indirect=True)
def test_update_pet(client):
    new_data = {'name': 'Ghost', 'category': 'dog'}

    rv = client.patch('/pets/1', json=new_data)
    assert rv.status_code == 200
    assert rv.json

    rv = client.get('/pets/1')
    assert rv.status_code == 200
    if client._example.startswith('base_response'):
        data = rv.json['data']
    else:
        data = rv.json
    assert data['name'] == new_data['name']
    assert data['category'] == new_data['category']

    rv = client.patch('/pets/13', json=new_data)
    assert rv.status_code == 404
    assert rv.json


@pytest.mark.parametrize('client', full_examples, indirect=True)
@pytest.mark.parametrize(
    'data', [{'name': 'Fyodor Mikhailovich Dostoevsky'}, {'category': 'human'}]
)
def test_update_pet_with_bad_data(client, data):
    rv = client.patch('/pets/1', json=data)
    assert rv.status_code == 422
    assert rv.json


@pytest.mark.skipif(flask_sqlalchemy_version < 3, reason='Need flask_sqlalchemy 3 for get_or_404')
@pytest.mark.parametrize('client', full_examples, indirect=True)
def test_delete_pet(client):
    rv = client.delete('/pets/1')
    assert rv.status_code == 204

    rv = client.get('/pets/1')
    assert rv.status_code == 404
    assert rv.json

    rv = client.delete('/pets/13')
    assert rv.status_code == 404
    assert rv.json


@pytest.mark.skipif(flask_sqlalchemy_version < 3, reason='Need flask_sqlalchemy 3 for paginate')
@pytest.mark.parametrize('client', ['pagination'], indirect=True)
def test_get_pets_pagination(client):
    rv = client.get('/pets')
    assert rv.status_code == 200
    assert rv.json
    assert 'pets' in rv.json
    assert 'pagination' in rv.json
    assert len(rv.json['pets']) == 20
    assert rv.json['pagination']['per_page'] == 20
    assert rv.json['pagination']['pages'] == 5
    assert rv.json['pagination']['total'] == 100
    assert rv.json['pagination']['page'] == 1
    assert rv.json['pets'][0]['id'] == 1
    assert rv.json['pets'][-1]['id'] == 20

    per_page = 10
    page = 2
    rv = client.get(f'/pets?per_page={per_page}&page={page}')
    assert rv.status_code == 200
    assert rv.json
    assert 'pets' in rv.json
    assert 'pagination' in rv.json
    assert len(rv.json['pets']) == per_page
    assert rv.json['pagination']['per_page'] == per_page
    assert rv.json['pagination']['pages'] == 10
    assert rv.json['pets'][0]['id'] == 11
    assert rv.json['pets'][-1]['id'] == 20


@pytest.mark.skipif(flask_sqlalchemy_version < 3, reason='Need flask_sqlalchemy 3 for paginate')
@pytest.mark.parametrize('client', ['pagination'], indirect=True)
def test_get_pets_pagination_with_bad_data(client):
    rv = client.get('/pets?per_page=100')
    assert rv.status_code == 422

    rv = client.get('/pets?page=100')
    assert rv.status_code == 404



from __future__ import annotations

import typing as t
from abc import ABC
from abc import abstractmethod

if t.TYPE_CHECKING:
    from flask import Request


class SchemaAdapter(ABC):
    """Base class for schema adapters.

    Schema adapters provide a common interface for different schema libraries
    (marshmallow, Pydantic, etc.) to integrate with APIFlask.

    *Version added: 3.0.0*
    """

    def __init__(self, schema: t.Any, schema_name: str | None = None, many: bool = False) -> None:
        """Initialize the adapter with a schema.

        Arguments:
            schema: The schema object (marshmallow Schema, Pydantic model, etc.)
            schema_name: Optional schema name (used by some adapters like marshmallow)
            many: Whether this schema represents a list/array of items
        """
        self.schema = schema
        self.many = many

    @abstractmethod
    def validate_input(self, request: Request, location: str, **kwargs: t.Any) -> t.Any:
        """Validate and parse input data from request.

        Arguments:
            request: Flask request object
            location: Location of data ('json', 'query', 'form', etc.)
            **kwargs: Additional arguments passed from decorator

        Returns:
            Validated and parsed data

        Raises:
            ValidationError: If validation fails
        """
        ...

    @abstractmethod
    def serialize_output(self, data: t.Any, many: bool = False) -> t.Any:
        """Serialize output data.

        Arguments:
            data: Data to serialize
            many: Whether to serialize many objects

        Returns:
            Serialized data ready for JSON response
        """
        ...

    @abstractmethod
    def get_openapi_schema(self, **kwargs: t.Any) -> dict[str, t.Any]:
        """Get OpenAPI schema definition.

        Arguments:
            **kwargs: Additional arguments for schema generation

        Returns:
            OpenAPI schema dict
        """
        ...

    @abstractmethod
    def get_schema_name(self) -> str:
        """Get the name of the schema for OpenAPI documentation.

        Returns:
            Schema name string
        """
        ...

    @property
    @abstractmethod
    def schema_type(self) -> str:
        """Get the type of schema adapter.

        Returns:
            Schema type identifier ('marshmallow', 'pydantic', etc.)
        """
        ...



import os

from werkzeug.utils import secure_filename
from apiflask import APIFlask, Schema
from apiflask.fields import File, String
from apiflask.validators import FileSize, FileType

app = APIFlask(__name__)

upload_dir = './upload'


class Image(Schema):
    image = File(validate=[FileType(['.png', '.jpg', '.jpeg', '.gif']), FileSize(max='5 MB')])


class ProfileIn(Schema):
    name = String()
    avatar = File(validate=[FileType(['.png', '.jpg', '.jpeg']), FileSize(max='2 MB')])


@app.post('/images')
@app.input(Image, location='files')
def upload_image(files_data):
    f = files_data['image']

    filename = secure_filename(f.filename)
    f.save(os.path.join(upload_dir, filename))

    return {'message': f'file {filename} saved.'}


@app.post('/profiles')
@app.input(ProfileIn, location='form_and_files')
def create_profile(form_and_files_data):
    avatar_file = form_and_files_data['avatar']
    name = form_and_files_data['name']

    avatar_filename = secure_filename(avatar_file.filename)
    avatar_file.save(os.path.join(upload_dir, avatar_filename))

    return {'message': f"{name}'s profile created."}



from typing import List
from enum import Enum

from apiflask import APIFlask, abort
from pydantic import BaseModel, Field

app = APIFlask(__name__)


class BaseResponse(BaseModel):
    data: object  # the data key
    message: str
    code: int


app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse
# the data key should match the data field name in the base response schema
# defaults to "data"
app.config['BASE_RESPONSE_DATA_KEY'] = 'data'

pets = [
    {'id': 0, 'name': 'Kitty', 'category': 'cat'},
    {'id': 1, 'name': 'Coco', 'category': 'dog'},
    {'id': 2, 'name': 'Flash', 'category': 'cat'},
]


class PetCategory(str, Enum):
    cat = 'cat'
    dog = 'dog'


class PetIn(BaseModel):
    name: str = Field(min_length=1, max_length=10)
    category: PetCategory


class PetOut(BaseModel):
    id: int
    name: str
    category: PetCategory


@app.get('/')
def say_hello():
    data = {'message': 'Hello!'}
    return {'data': data, 'message': 'Success!', 'code': 200}


@app.get('/pets/<int:pet_id>')
@app.output(PetOut)
def get_pet(pet_id):
    if pet_id > len(pets) - 1 or pets[pet_id].get('deleted'):
        abort(404)
    return {
        'data': pets[pet_id],
        'message': 'Success!',
        'code': 200,
    }


@app.get('/pets')
@app.output(List[PetOut])
def get_pets():
    return {
        'data': pets,
        'message': 'Success!',
        'code': 200,
    }


@app.post('/pets')
@app.input(PetIn, location='json')
@app.output(PetOut, status_code=201)
def create_pet(json_data):
    new_id = len(pets) + 1
    new_pet = PetOut(id=new_id, name=json_data.name, category=json_data.category)
    pets.append(new_pet)
    return {'data': new_pet, 'message': 'Pet created.', 'code': 201}


@app.patch('/pets/<int:pet_id>')
@app.input(PetIn, location='json')
@app.output(PetOut)
def update_pet(pet_id, json_data):
    if pet_id > len(pets) - 1:
        abort(404)

    pet = pets[pet_id]
    pet['name'] = json_data.name
    pet['category'] = json_data.category
    return {'data': pet, 'message': 'Pet updated.', 'code': 200}


@app.delete('/pets/<int:pet_id>')
@app.output({}, status_code=204)
def delete_pet(pet_id):
    if pet_id > len(pets) - 1:
        abort(404)
    pets[pet_id]['deleted'] = True
    pets[pet_id]['name'] = 'Ghost'
    return ''



from dataclasses import field

from marshmallow_dataclass import dataclass
from apiflask import APIFlask, abort
from apiflask.validators import Length, OneOf

app = APIFlask(__name__)


@dataclass
class PetIn:
    name: str = field(
        metadata={
            'required': True,
            'validate': Length(min=1, max=10),
            'metadata': {
                'example': 'Medor',
                'description': 'This will be printed in the generated doc. '
                'The "example" value "Medor" will be fed '
                'into the "try it"/"Send API request".',
            },
        }
    )
    category: str = field(
        default='dog', metadata={'required': True, 'validate': OneOf(['cat', 'dog'])}
    )


@dataclass
class PetOut:
    id: int
    name: str
    category: str


pets = [
    {'id': 0, 'name': 'Kitty', 'category': 'cat'},
    {'id': 1, 'name': 'Coco', 'category': 'dog'},
    {'id': 2, 'name': 'Flash', 'category': 'cat'},
]


@app.get('/')
def say_hello():
    return {'message': 'Hello!'}


@app.get('/pets/<int:pet_id>')
@app.output(PetOut.Schema)
def get_pet(pet_id):
    if pet_id > len(pets) - 1 or pets[pet_id].get('deleted'):
        abort(404)
    return pets[pet_id]


@app.get('/pets')
@app.output(PetOut.Schema(many=True))
def get_pets():
    return pets


@app.post('/pets')
@app.input(PetIn.Schema, location='json', arg_name='pet')
@app.output(PetOut.Schema, status_code=201)
def create_pet(pet: PetIn):
    pet_id = len(pets)
    pets.append({'id': pet_id, 'name': pet.name, 'category': pet.category})
    return pets[pet_id]


# partial=True is not supported in marshmallow-dataclass currently
# https://github.com/lovasoa/marshmallow_dataclass/issues/169
@app.patch('/pets/<int:pet_id>')
@app.input(PetIn.Schema, location='json', arg_name='pet')
@app.output(PetOut.Schema)
def update_pet(pet_id, pet: PetIn):
    if pet_id > len(pets) - 1:
        abort(404)
    pets[pet_id]['name'] = pet.name
    pets[pet_id]['category'] = pet.category
    return pets[pet_id]


@app.delete('/pets/<int:pet_id>')
@app.output({}, status_code=204)
def delete_pet(pet_id):
    if pet_id > len(pets) - 1:
        abort(404)
    pets[pet_id]['deleted'] = True
    pets[pet_id]['name'] = 'Ghost'
    return ''



from __future__ import annotations

import typing as t
from io import IOBase

from flask import current_app

from ..exceptions import _ValidationError
from .base import SchemaAdapter

if t.TYPE_CHECKING:
    from flask import Request

try:
    from pydantic import BaseModel, ValidationError as PydanticValidationError
    from pydantic.fields import FieldInfo
    from pydantic_core import ErrorDetails

    HAS_PYDANTIC = True
except ImportError:
    BaseModel = None  # type: ignore
    PydanticValidationError = None  # type: ignore
    FieldInfo = None  # type: ignore
    ErrorDetails = None  # type: ignore
    HAS_PYDANTIC = False


def _format_pydantic_errors(errors: list[ErrorDetails]) -> dict[str, list[str]]:
    """Format Pydantic validation errors to match marshmallow format.

    Arguments:
        errors: List of Pydantic error dicts

    Returns:
        Dict mapping field names to error message lists
    """
    formatted_errors: dict[str, list[str]] = {}

    for error in errors:
        # Get the field path
        field_path = '.'.join(str(loc) for loc in error['loc'])
        if not field_path:
            field_path = '_schema'

        # Get the error message
        message = error['msg']

        if field_path not in formatted_errors:
            formatted_errors[field_path] = []
        formatted_errors[field_path].append(message)

    return formatted_errors


class PydanticAdapter(SchemaAdapter):
    """Schema adapter for Pydantic models."""

    def __init__(
        self,
        schema: type[BaseModel] | BaseModel,
        schema_name: str | None = None,
        many: bool = False,
    ) -> None:
        """Initialize the Pydantic adapter.

        Arguments:
            schema: Pydantic model class or instance
            schema_name: Optional schema name (not used for Pydantic)
            many: Whether this schema represents a list/array of items
        """
        if not HAS_PYDANTIC:
            raise ImportError(
                'Pydantic is required for PydanticAdapter. ' 'Install it with: pip install pydantic'
            )

        if isinstance(schema, type) and issubclass(schema, BaseModel):
            self.model_class = schema
            self.schema = schema
        elif isinstance(schema, BaseModel):
            self.model_class = schema.__class__
            self.schema = schema.__class__
        else:
            raise TypeError(f'Expected Pydantic model, got {type(schema)}')

        self.many = many

    @property
    def schema_type(self) -> str:
        return 'pydantic'

    def validate_input(self, request: Request, location: str, **kwargs: t.Any) -> BaseModel:
        """Validate input using Pydantic."""
        try:
            if location == 'json':
                data = request.get_json(force=True)
                if data is None:
                    data = {}
                return self.model_class.model_validate(data)

            elif location == 'query' or location == 'querystring':
                # Handle query parameters
                data = request.args.to_dict()
                return self.model_class.model_validate(data)

            elif location == 'form':
                # Handle form data
                data = request.form.to_dict()
                return self.model_class.model_validate(data)

            elif location == 'files':
                # Handle file uploads with form data
                data = {}
                data.update(request.form.to_dict())

                # Add files to data
                for key, file in request.files.items():
                    if isinstance(file, IOBase):
                        data[key] = file

                return self.model_class.model_validate(data)

            elif location == 'form_and_files':
                # Combine form and files
                data = request.form.to_dict()
                for key, file in request.files.items():
                    data[key] = file  # type: ignore
                return self.model_class.model_validate(data)

            elif location == 'json_or_form':
                # Try JSON first, then form
                if request.is_json:
                    data = request.get_json(force=True) or {}
                else:
                    data = request.form.to_dict()
                return self.model_class.model_validate(data)

            elif location == 'cookies':
                # Handle cookies
                data = request.cookies.to_dict()
                return self.model_class.model_validate(data)

            elif location == 'headers':
                # Handle headers - convert header names to field names
                # HTTP headers like X-Token become x_token for Pydantic fields
                data = {}
                for header_name, value in request.headers:
                    # Convert header name to field name (e.g., X-Token -> x_token)
                    field_name = header_name.lower().replace('-', '_')
                    data[field_name] = value
                return self.model_class.model_validate(data)

            elif location == 'path' or location == 'view_args':
                # Handle path/view_args
                data = request.view_args or {}
                return self.model_class.model_validate(data)

            else:
                raise ValueError(f'Unsupported location: {location}')

        except PydanticValidationError as error:
            formatted_errors = _format_pydantic_errors(error.errors())
            raise _ValidationError(
                current_app.config['VALIDATION_ERROR_STATUS_CODE'],
                current_app.config['VALIDATION_ERROR_DESCRIPTION'],
                {location: formatted_errors},
            ) from error

    def serialize_output(self, data: t.Any, many: bool = False) -> t.Any:
        """Serialize output using Pydantic with validation."""
        if many and isinstance(data, (list, tuple)):
            # Handle lists of data
            result = []
            for item in data:
                if isinstance(item, BaseModel):
                    # Already validated, just serialize
                    result.append(item.model_dump(mode='json'))
                else:
                    # Validate and serialize
                    validated = self.model_class.model_validate(item)
                    result.append(validated.model_dump(mode='json'))
            return result
        elif isinstance(data, BaseModel):
            # Pydantic model instance - already validated, just serialize
            return data.model_dump(mode='json')
        elif isinstance(data, (list, tuple)) and not many:
            # Handle lists when many=False
            result = []
            for item in data:
                if isinstance(item, BaseModel):
                    result.append(item.model_dump(mode='json'))
                else:
                    # Validate and serialize
                    validated = self.model_class.model_validate(item)
                    result.append(validated.model_dump(mode='json'))
            return result
        else:
            # Validate and serialize (dicts, primitives)
            validated = self.model_class.model_validate(data)
            return validated.model_dump(mode='json')

    def get_openapi_schema(self, **kwargs: t.Any) -> dict[str, t.Any]:
        """Get OpenAPI schema from Pydantic model.

        Uses Pydantic's ref_template to generate OpenAPI-compliant $ref paths.
        Nested models in $defs should be extracted and registered separately
        at the components/schemas level.
        """
        try:
            # Generate schema name for use in ref_template
            schema_name = self.get_schema_name()

            # Use ref_template to make Pydantic generate OpenAPI-compatible refs
            # The {model} placeholder will be replaced with nested model names
            ref_template = f'#/components/schemas/{schema_name}.{{model}}'

            # Generate JSON schema with OpenAPI-compatible references
            schema = self.model_class.model_json_schema(
                ref_template=ref_template, mode=kwargs.get('mode', 'validation')
            )

            # Note: The $defs should be extracted and registered separately
            # in components/schemas. The caller (app.py) should handle this.

            return schema

        except Exception:
            # Fallback to basic schema
            return {'type': 'object', 'title': self.get_schema_name()}

    def get_schema_name(self) -> str:
        """Get schema name for OpenAPI documentation."""
        return self.model_class.__name__



import pytest

from apiflask import APIFlask


def test_docs_favicon(app, client):
    app.config['DOCS_FAVICON'] = '/my-favicon.png'

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'href="/my-favicon.png"' in rv.data


@pytest.mark.parametrize('config_value', [True, False])
def test_docs_use_google_font(client, config_value):
    app = APIFlask(__name__, docs_ui='redoc')
    app.config['REDOC_USE_GOOGLE_FONT'] = config_value
    client = app.test_client()

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert bool(b'fonts.googleapis.com' in rv.data) is config_value


def test_redoc_standalone_js(client):
    app = APIFlask(__name__, docs_ui='redoc')
    app.config['REDOC_STANDALONE_JS'] = 'https://cdn.example.com/redoc.js'
    client = app.test_client()

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'src="https://cdn.example.com/redoc.js"' in rv.data


@pytest.mark.parametrize('config_value', [{}, {'disableSearch': True, 'hideLoading': True}])
def test_redoc_config(client, config_value):
    app = APIFlask(__name__, docs_ui='redoc')
    app.config['REDOC_CONFIG'] = config_value
    client = app.test_client()

    rv = client.get('/docs')
    assert rv.status_code == 200
    if config_value == {}:
        assert b'{},' in rv.data
    else:
        assert b'"disableSearch": true' in rv.data
        assert b'"hideLoading": true' in rv.data


def test_swagger_ui_resources(app, client):
    app.config['SWAGGER_UI_CSS'] = 'https://cdn.example.com/swagger-ui.css'
    app.config['SWAGGER_UI_BUNDLE_JS'] = 'https://cdn.example.com/swagger-ui.bundle.js'
    app.config['SWAGGER_UI_STANDALONE_PRESET_JS'] = 'https://cdn.example.com/swagger-ui.preset.js'

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'href="https://cdn.example.com/swagger-ui.css"' in rv.data
    assert b'src="https://cdn.example.com/swagger-ui.bundle.js"' in rv.data
    assert b'src="https://cdn.example.com/swagger-ui.preset.js"' in rv.data


def test_swagger_ui_layout(app, client):
    app.config['SWAGGER_UI_LAYOUT'] = 'StandaloneLayout'

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'StandaloneLayout' in rv.data
    assert b'BaseLayout' not in rv.data


def test_swagger_ui_config(app, client):
    app.config['SWAGGER_UI_CONFIG'] = {'deepLinking': False, 'layout': 'StandaloneLayout'}

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'"deepLinking": false' in rv.data
    assert b'"layout": "StandaloneLayout"' in rv.data


def test_swagger_ui_oauth_config(app, client):
    app.config['SWAGGER_UI_OAUTH_CONFIG'] = {
        'clientId': 'foo',
        'usePkceWithAuthorizationCodeGrant': True,
    }

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'ui.initOAuth(' in rv.data
    assert b'"clientId": "foo"' in rv.data
    assert b'"usePkceWithAuthorizationCodeGrant": true' in rv.data


def test_elements_config():
    app = APIFlask(__name__, docs_ui='elements')

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    # test default router
    assert b'router="hash"' in rv.data

    app.config['ELEMENTS_CONFIG'] = {'hideTryIt': False, 'router': 'memory'}

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'hideTryIt=false' in rv.data
    assert b'router="memory"' in rv.data


def test_elements_layout():
    app = APIFlask(__name__, docs_ui='elements')
    app.config['ELEMENTS_LAYOUT'] = 'stacked'

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'layout="stacked"' in rv.data
    assert b'layout="sidebar"' not in rv.data


def test_elements_resources():
    app = APIFlask(__name__, docs_ui='elements')
    app.config['ELEMENTS_CSS'] = 'https://cdn.example.com/elements.css'
    app.config['ELEMENTS_JS'] = 'https://cdn.example.com/elements.js'

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'href="https://cdn.example.com/elements.css"' in rv.data
    assert b'src="https://cdn.example.com/elements.js"' in rv.data


def test_rapidoc_config():
    app = APIFlask(__name__, docs_ui='rapidoc')
    app.config['RAPIDOC_CONFIG'] = {'update-route': False, 'layout': 'row'}

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'update-route=false' in rv.data
    assert b'layout="row"' in rv.data


def test_rapidoc_theme():
    app = APIFlask(__name__, docs_ui='rapidoc')
    app.config['RAPIDOC_THEME'] = 'dark'

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'theme="dark"' in rv.data
    assert b'theme="light"' not in rv.data


def test_rapidoc_resources():
    app = APIFlask(__name__, docs_ui='rapidoc')
    app.config['RAPIDOC_JS'] = 'https://cdn.example.com/rapidoc.js'

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'src="https://cdn.example.com/rapidoc.js"' in rv.data


def test_rapipdf_config():
    app = APIFlask(__name__, docs_ui='rapipdf')
    app.config['RAPIPDF_CONFIG'] = {'include-example': True, 'button-label': 'Generate!'}

    rv = app.test_client().get('/docs')
    assert rv.status_code == 200
    assert b'include-example=true' in rv.data
    assert b'button-label="Generate!"' in rv.data


def test_rapipdf_resources():
    app = APIFlask(__name__, docs_ui='rapipdf')
    client = app.test_client()
    app.config['RAPIPDF_JS'] = 'https://cdn.example.com/rapipdf.js'

    rv = client.get('/docs')
    assert rv.status_code == 200
    assert b'src="https://cdn.example.com/rapipdf.js"' in rv.data



from apiflask import APIFlask
from opentelemetry.instrumentation.flask import FlaskInstrumentor
from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.semconv.resource import ResourceAttributes

app = APIFlask(__name__, title='apiflask-otel', version='1.0.0')


def configure_trace(app: APIFlask):
    """initialize OpenTelemetry based on the provided settings.

    Args:
        app (APIFlask): APIFlask instance
    """
    resource = Resource(
        attributes={
            ResourceAttributes.SERVICE_NAME: app.title,
            ResourceAttributes.SERVICE_VERSION: app.version,
        }
    )
    tracer_provider = TracerProvider(resource=resource)
    trace.set_tracer_provider(tracer_provider)

    FlaskInstrumentor().instrument_app(app)


configure_trace(app=app)


@app.get('/')
def index():
    span_ctx = trace.get_current_span().get_span_context()
    print(f'trace id: {span_ctx.trace_id}, span id: {span_ctx.span_id}')
    return {'trace_id': span_ctx.trace_id, 'span_id': span_ctx.span_id}



from __future__ import annotations

import typing as t

from .schemas import http_error_schema
from .schemas import validation_error_schema
from .types import OpenAPISchemaType
from .types import TagsType


# OpenAPI fields
OPENAPI_VERSION: str = '3.0.3'
SERVERS: list[dict[str, str]] | None = None
TAGS: TagsType | None = None
EXTERNAL_DOCS: dict[str, str] | None = None
INFO: dict[str, str | dict] | None = None
DESCRIPTION: str | None = None
TERMS_OF_SERVICE: str | None = None
CONTACT: dict[str, str] | None = None
LICENSE: dict[str, str] | None = None
SECURITY_SCHEMES: dict[str, t.Any] | None = None
# OpenAPI spec
SPEC_FORMAT: str = 'json'
YAML_SPEC_MIMETYPE: str = 'text/vnd.yaml'
JSON_SPEC_MIMETYPE: str = 'application/json'
LOCAL_SPEC_PATH: str | None = None
LOCAL_SPEC_JSON_INDENT: int = 2
SYNC_LOCAL_SPEC: bool | None = None
SPEC_PROCESSOR_PASS_OBJECT: bool = False
SPEC_DECORATORS: list[t.Callable] | None = None
DOCS_DECORATORS: list[t.Callable] | None = None
SWAGGER_UI_OAUTH_REDIRECT_DECORATORS: list[t.Callable] | None = None
# Automation behavior control
AUTO_TAGS: bool = True
AUTO_SERVERS: bool = True
AUTO_OPERATION_SUMMARY: bool = True
AUTO_OPERATION_DESCRIPTION: bool = True
AUTO_OPERATION_ID: bool = False
AUTO_200_RESPONSE: bool = True
AUTO_404_RESPONSE: bool = True
AUTO_VALIDATION_ERROR_RESPONSE: bool = True
AUTO_AUTH_ERROR_RESPONSE: bool = True
# Response customization
SUCCESS_DESCRIPTION: str = 'Successful response'
NOT_FOUND_DESCRIPTION: str = 'Not found'
VALIDATION_ERROR_DESCRIPTION: str = 'Validation error'
AUTH_ERROR_DESCRIPTION: str = 'Authentication error'
VALIDATION_ERROR_STATUS_CODE: int = 422
AUTH_ERROR_STATUS_CODE: int = 401
VALIDATION_ERROR_SCHEMA: OpenAPISchemaType = validation_error_schema
HTTP_ERROR_SCHEMA: OpenAPISchemaType = http_error_schema
BASE_RESPONSE_SCHEMA: OpenAPISchemaType | None = None
BASE_RESPONSE_DATA_KEY: str = 'data'
# API docs
DOCS_FAVICON: str = 'https://apiflask.com/_assets/favicon.png'
REDOC_USE_GOOGLE_FONT: bool = True
REDOC_STANDALONE_JS: str = 'https://cdn.redoc.ly/redoc/latest/bundles/\
redoc.standalone.js'  # TODO: rename to REDOC_JS
REDOC_CONFIG: dict | None = None
SWAGGER_UI_CSS: str = 'https://cdn.jsdelivr.net/npm/swagger-ui-dist/swagger-ui.css'
SWAGGER_UI_BUNDLE_JS: str = 'https://cdn.jsdelivr.net/npm/swagger-ui-dist/\
swagger-ui-bundle.js'  # TODO: rename to SWAGGER_UI_JS
SWAGGER_UI_STANDALONE_PRESET_JS: str = 'https://cdn.jsdelivr.net/npm/swagger-ui-dist/\
swagger-ui-standalone-preset.js'  # TODO: rename to SWAGGER_UI_STANDALONE_JS
SWAGGER_UI_LAYOUT: str = 'BaseLayout'
SWAGGER_UI_CONFIG: dict | None = None
SWAGGER_UI_OAUTH_CONFIG: dict | None = None
ELEMENTS_JS: str = 'https://cdn.jsdelivr.net/npm/@stoplight/elements/web-components.min.js'
ELEMENTS_CSS: str = 'https://cdn.jsdelivr.net/npm/@stoplight/elements/styles.min.css'
ELEMENTS_LAYOUT: str = 'sidebar'
ELEMENTS_CONFIG: dict | None = None
RAPIDOC_JS: str = 'https://cdn.jsdelivr.net/npm/rapidoc/dist/rapidoc-min.js'
RAPIDOC_THEME: str = 'light'
RAPIDOC_CONFIG: dict | None = None
RAPIPDF_JS: str = 'https://cdn.jsdelivr.net/npm/rapipdf/dist/rapipdf-min.js'
RAPIPDF_CONFIG: dict | None = None

# Version changed: 1.2.0
# Change VALIDATION_ERROR_STATUS_CODE from 400 to 422.

# Version added: 1.3.0
# SPEC_PROCESSOR_PASS_OBJECT



from __future__ import annotations

import typing as t
import warnings

from flask import current_app
from flask import g
from flask import request
from flask_httpauth import HTTPBasicAuth as BaseHTTPBasicAuth
from flask_httpauth import HTTPTokenAuth as BaseHTTPTokenAuth
from flask_httpauth import MultiAuth as BaseMultiAuth
from werkzeug.datastructures import Authorization

from .exceptions import HTTPError
from .types import ErrorCallbackType
from .types import HTTPAuthType
from .types import ResponseReturnValueType
from .types import SecurityScheme


class _AuthBase:
    """Base class for `HTTPBasicAuth` and `HTTPBasicAuth`."""

    def __init__(
        self,
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        self.description = description
        self.security_scheme_name = security_scheme_name
        self.error_handler(self._auth_error_handler)  # type: ignore

    @property
    def current_user(self) -> None | t.Any:
        return g.get('flask_httpauth_user', None)

    @staticmethod
    def _auth_error_handler(status_code: int) -> ResponseReturnValueType:
        """The default error handler for Flask-HTTPAuth.

        This handler will return JSON response when setting `APIFlask(json_errors=True)` (default).

        *Version changed: 0.9.0*

        - The default reason phrase is used for auth errors.
        - It will call the `app.error_callback` for auth errors.
        """
        error = HTTPError(status_code)
        if current_app.json_errors:  # type: ignore
            return current_app.error_callback(error)  # type: ignore
        return error.message, status_code  # type: ignore

    def error_processor(self, f: ErrorCallbackType) -> None:
        """A decorator to register an error callback function for auth errors (401/403).

        The error callback function will be called when authentication errors happened.
        It should accept an `HTTPError` instance and return a valid response. APIFlask will pass
        the callback function you decorated to Flask-HTTPAuth's `error_handler` method internally.

        Example:

        ```python
        from apiflask import APIFlask, HTTPTokenAuth

        app = APIFlask(__name__)
        auth = HTTPTokenAuth()

        @auth.error_processor
        def my_auth_error_processor(error):
            return {
                'status_code': error.status_code,
                'message': error.message
            }, error.status_code
        ```

        See more details of the error object in
        [APIFlask.error_processor][apiflask.APIFlask.error_processor].

        *Version added: 0.9.0*
        """
        self.error_handler(lambda status_code: f(HTTPError(status_code)))  # type: ignore


class HTTPBasicAuth(_AuthBase, BaseHTTPBasicAuth, SecurityScheme):
    """Flask-HTTPAuth's HTTPBasicAuth with some modifications.

    - Add an authentication error handler that returns JSON response.
    - Expose the `auth.current_user` as a property.
    - Add a `description` attribute for OpenAPI Spec.
    - Add a `name` attribute for OpenAPI Spec.
    - Add the `get_security_scheme` method for OpenAPI Spec.

    Examples:

    ```python
    from apiflask import APIFlask, HTTPBasicAuth

    app = APIFlask(__name__)
    auth = HTTPBasicAuth()
    ```

    *Version changed: 3.0.0*

    - Add parameter `name`.

    *Version changed: 1.3.0*

    - Add `security_scheme_name` parameter.
    """

    def __init__(
        self,
        name: str = 'BasicAuth',
        scheme: str = 'Basic',
        realm: str | None = None,
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        """Initialize an `HTTPBasicAuth` object.

        Arguments:
            name: The security scheme name, default to `BasicAuth`.
            scheme: The authentication scheme used in the `WWW-Authenticate`
                header. Defaults to `'Basic'`.
            realm: The realm used in the `WWW-Authenticate` header to indicate
                a scope of protection, defaults to use `'Authentication Required'`.
            description: The description of the OpenAPI security scheme.
            security_scheme_name: The name of the OpenAPI security scheme, default to `BasicAuth`.
        """
        BaseHTTPBasicAuth.__init__(self, scheme=scheme, realm=realm)
        super().__init__(description=description, security_scheme_name=security_scheme_name)
        self.name = security_scheme_name or name

    def get_security_scheme(self) -> dict[str, t.Any]:
        security_scheme = {
            'type': 'http',
            'scheme': 'basic',
        }

        if self.description is not None:
            security_scheme['description'] = self.description

        return security_scheme


class HTTPTokenAuth(_AuthBase, BaseHTTPTokenAuth, SecurityScheme):
    """Flask-HTTPAuth's HTTPTokenAuth with some modifications.

    - Add an authentication error handler that returns JSON response.
    - Expose the `auth.current_user` as a property.
    - Add a `description` attribute for OpenAPI Spec.
    - Add a `name` attribute for OpenAPI Spec.
    - Add the `get_security_scheme` method for OpenAPI Spec.

    Examples:

    ```python
    from apiflask import APIFlask, HTTPTokenAuth

    app = APIFlask(__name__)
    auth = HTTPTokenAuth()
    ```
    """

    def __init__(
        self,
        scheme: str = 'Bearer',
        name: str = 'BearerAuth',
        realm: str | None = None,
        header: str | None = None,
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        """Initialize a `HTTPTokenAuth` object.

        Arguments:
            scheme: The authentication scheme used in the `WWW-Authenticate`
                header. One of `'Bearer'` and `'ApiKey'`, defaults to `'Bearer'`.
            name: The security scheme name, default to `BearerAuth`.
            realm: The realm used in the `WWW-Authenticate` header to indicate
                 a scope of protection, defaults to use `'Authentication Required'`.
            header: The custom header where to obtain the token (instead
                of from `Authorization` header). If a custom header is used,
                the scheme should not be included. Example:

                ```
                X-API-Key: this-is-my-token
                ```

            description: The description of the OpenAPI security scheme.
            security_scheme_name: The name of the OpenAPI security scheme,
                defaults to `BearerAuth` or `ApiKeyAuth`.

        *Version changed: 3.0.0*

        - Add parameter `name`.

        *Version changed: 1.3.0*

        - Add `security_scheme_name` parameter.
        """
        BaseHTTPTokenAuth.__init__(self, scheme=scheme, realm=realm, header=header)
        super().__init__(description=description, security_scheme_name=security_scheme_name)
        if not self.scheme.lower() == 'bearer' or self.header is not None:
            name = 'ApiKeyAuth'
            warnings.warn(
                'The API key authorization by HTTPTokenAuth is deprecated and will be removed in '
                'APIFlask 4.0.0. Use '
                'APIKeyHeaderAuth, APIKeyCookieAuth, or APIKeyQueryAuth instead.',
                DeprecationWarning,
                stacklevel=2,
            )

        self.name = security_scheme_name or name

    def get_security_scheme(self) -> dict[str, t.Any]:
        if not self.scheme.lower() == 'bearer' or self.header is not None:
            security_scheme = {
                'type': 'apiKey',
                'name': self.header,
                'in': 'header',
            }
        else:
            security_scheme = {
                'type': 'http',
                'scheme': 'bearer',
            }

        if self.description is not None:
            security_scheme['description'] = self.description

        return security_scheme


class _BaseAPIKeyAuth(_AuthBase, BaseHTTPTokenAuth):
    """Base class for `APIKeyHeaderAuth` `APIKeyCookieAuth` `APIKeyQueryAuth`."""

    def __init__(
        self,
        name: str = 'ApiKeyAuth',
        scheme: str = 'ApiKey',
        realm: str | None = None,
        param_name: str | None = 'X-API-Key',
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        BaseHTTPTokenAuth.__init__(self, scheme=scheme, realm=realm, header=param_name)
        super().__init__(description=description, security_scheme_name=security_scheme_name)
        self.name = security_scheme_name or name
        self.param_name = param_name
        self._location = 'header'

    def get_security_scheme(self) -> dict[str, t.Any]:
        security_scheme = {
            'type': 'apiKey',
            'name': self.param_name,
            'in': self._location,
        }

        if self.description is not None:
            security_scheme['description'] = self.description

        return security_scheme


class APIKeyHeaderAuth(_BaseAPIKeyAuth):
    """Flask-HTTPAuth's HTTPTokenAuth with some modifications to implement header APIKey authentication.

    - Add an authentication error handler that returns JSON response.
    - Expose the `auth.current_user` as a property.
    - Add a `description` attribute for OpenAPI Spec.
    - Add a `name` attribute for OpenAPI Spec.
    - Add a `param_name` attribute for API Key paramter.
    - Add the `get_security_scheme` method for OpenAPI Spec.

    Examples:

    ```python
    from apiflask import APIFlask, APIKeyHeaderAuth

    app = APIFlask(__name__)
    auth = APIKeyHeaderAuth()
    ```
    """  # noqa: E501

    def __init__(
        self,
        name: str = 'ApiKeyAuth',
        scheme: str = 'ApiKey',
        realm: str | None = None,
        param_name: str | None = 'X-API-Key',
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        """Initialize a `APIKeyHeaderAuth` object.

        Arguments:
            name: The security scheme name, default to `ApiKeyAuth`.
            scheme: The authentication scheme used in the `WWW-Authenticate`
                header. defaults to `'ApiKey'`.
            realm: The realm used in the `WWW-Authenticate` header to indicate
                 a scope of protection, defaults to use `'Authentication Required'`.
            param_name: The name of API Key paramter. defaults to `'Authorization'`.
            description: The description of the OpenAPI security scheme.
            security_scheme_name: The name of the OpenAPI security scheme,
                defaults to `ApiKeyAuth`.

        *Version added: 3.0.0*
        """
        super().__init__(
            name=name,
            scheme=scheme,
            realm=realm,
            param_name=param_name,
            description=description,
            security_scheme_name=security_scheme_name,
        )


class APIKeyCookieAuth(_BaseAPIKeyAuth):
    """Flask-HTTPAuth's HTTPTokenAuth with some modifications to implement cookie APIKey authentication.

    - Add an authentication error handler that returns JSON response.
    - Expose the `auth.current_user` as a property.
    - Add a `description` attribute for OpenAPI Spec.
    - Add a `name` attribute for OpenAPI Spec.
    - Add a `param_name` attribute for API Key paramter.
    - Add the `get_security_scheme` method for OpenAPI Spec.

    Examples:

    ```python
    from apiflask import APIFlask, APIKeyCookieAuth

    app = APIFlask(__name__)
    auth = APIKeyCookieAuth()
    ```
    """  # noqa: E501

    def __init__(
        self,
        name: str = 'ApiKeyAuth',
        scheme: str = 'ApiKey',
        realm: str | None = None,
        param_name: str | None = 'X-API-Key',
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        """Initialize a `APIKeyCookieAuth` object.

        Arguments:
            name: The security scheme name, default to `ApiKeyAuth`.
            scheme: The authentication scheme used in the `WWW-Authenticate`
                header. defaults to `'ApiKey'`.
            realm: The realm used in the `WWW-Authenticate` header to indicate
                 a scope of protection, defaults to use `'Authentication Required'`.
            param_name: The name of API Key paramter. defaults to `'ApiKey'`.
            description: The description of the OpenAPI security scheme.
            security_scheme_name: The name of the OpenAPI security scheme,
                defaults to `ApiKeyAuth`.

        *Version added: 3.0.0*
        """
        super().__init__(
            name=name,
            scheme=scheme,
            realm=realm,
            param_name=param_name,
            description=description,
            security_scheme_name=security_scheme_name,
        )
        self._location = 'cookie'

    def is_compatible_auth(self, headers):
        return self.param_name in request.cookies

    def get_auth(self) -> Authorization:
        auth = Authorization('apiKey')
        auth.token = request.cookies.get(self.param_name, '')  # type: ignore
        return auth


class APIKeyQueryAuth(_BaseAPIKeyAuth):
    """Flask-HTTPAuth's HTTPTokenAuth with some modifications to implement query parameter APIKey authentication.

    - Add an authentication error handler that returns JSON response.
    - Expose the `auth.current_user` as a property.
    - Add a `description` attribute for OpenAPI Spec.
    - Add a `name` attribute for OpenAPI Spec.
    - Add a `param_name` attribute for API Key paramter.
    - Add the `get_security_scheme` method for OpenAPI Spec.

    Examples:

    ```python
    from apiflask import APIFlask, APIKeyQueryAuth

    app = APIFlask(__name__)
    auth = APIKeyQueryAuth()
    ```
    """  # noqa: E501

    def __init__(
        self,
        name: str = 'ApiKeyAuth',
        scheme: str = 'ApiKey',
        realm: str | None = None,
        param_name: str | None = 'X-API-Key',
        description: str | None = None,
        security_scheme_name: str | None = None,
    ) -> None:
        """Initialize a `APIKeyQueryAuth` object.

        Arguments:
            name: The security scheme name, default to `ApiKeyAuth`.
            scheme: The authentication scheme used in the `WWW-Authenticate`
                header. defaults to `'ApiKey'`.
            realm: The realm used in the `WWW-Authenticate` header to indicate
                 a scope of protection, defaults to use `'Authentication Required'`.
            param_name: The name of API Key paramter. defaults to `'ApiKey'`.
            description: The description of the OpenAPI security scheme.
            security_scheme_name: The name of the OpenAPI security scheme,
                defaults to `ApiKeyAuth`.

        *Version added: 3.0.0*
        """
        super().__init__(
            name=name,
            scheme=scheme,
            realm=realm,
            param_name=param_name,
            description=description,
            security_scheme_name=security_scheme_name,
        )
        self._location = 'query'

    def is_compatible_auth(self, headers):
        return self.param_name in request.args

    def get_auth(self) -> Authorization:
        auth = Authorization('apiKey')
        auth.token = request.args.get(self.param_name, '')  # type: ignore
        return auth


class MultiAuth(BaseMultiAuth):
    """Flask-HTTPAuth's HTTPMultiAuth with some modifications.

    - Expose the `auth.current_user` as a property.

    Examples:

    ```python
    from apiflask import APIFlask, HTTPBasicAuth, HTTPTokenAuth, MultiAuth

    app = APIFlask(__name__)
    basic_auth = HTTPBasicAuth()
    token_auth = HTTPTokenAuth()
    multi_auth = MultiAuth(basic_auth, token_auth)
    ```

    *Version added: 3.0.0*
    """

    def __init__(self, main_auth: HTTPAuthType, *additional_auth: tuple[HTTPAuthType]):
        """Initialize a `HTTPMultiAuth` object.

        Arguments:
            main_auth: The main authentication object.
            additional_auth: The additional additional objects.
        """
        super().__init__(main_auth, *additional_auth)

    @property
    def current_user(self) -> None | t.Any:
        return g.get('flask_httpauth_user', None)

    @property
    def _auths(self) -> list[HTTPAuthType]:
        return [self.main_auth] + list(self.additional_auth)



from __future__ import annotations

import typing as t

from flask import current_app
from marshmallow import Schema
from marshmallow import ValidationError as MarshmallowValidationError
from webargs.flaskparser import FlaskParser as BaseFlaskParser
from webargs.multidictproxy import MultiDictProxy

from ..exceptions import _ValidationError
from ..schemas import EmptySchema
from ..schemas import FileSchema
from .base import SchemaAdapter

if t.TYPE_CHECKING:
    from flask import Request


BODY_LOCATIONS = ['json', 'files', 'form', 'form_and_files', 'json_or_form']


class FlaskParser(BaseFlaskParser):
    """Overwrite the default `webargs.FlaskParser.handle_error`.

    Update the default status code and the error description from related
    configuration variables.

    *Version added: 3.0.0*
    """

    USE_ARGS_POSITIONAL = False

    def handle_error(  # type: ignore
        self,
        error: MarshmallowValidationError,
        req: Request,
        schema: Schema,
        *,
        error_status_code: int,
        error_headers: t.Mapping[str, str],
    ) -> None:
        raise _ValidationError(
            error_status_code or current_app.config['VALIDATION_ERROR_STATUS_CODE'],
            current_app.config['VALIDATION_ERROR_DESCRIPTION'],
            error.messages,
            error_headers,
        )

    def load_location_data(self, *, schema: Schema, req: Request, location: str) -> t.Any:
        return self._load_location_data(schema=schema, req=req, location=location)


parser: FlaskParser = FlaskParser()


@parser.location_loader('form_and_files')
def load_form_and_files(request: Request, schema: t.Any) -> MultiDictProxy:
    return _get_files_and_form(request, schema)


@parser.location_loader('files')
def load_files(request: Request, schema: t.Any) -> MultiDictProxy:
    return _get_files_and_form(request, schema)


def _get_files_and_form(request: Request, schema: t.Any) -> MultiDictProxy:
    """Get files and form data and format field values with schema."""
    form_and_files_data = request.files.copy()
    # Type ignore needed due to MultiDict type incompatibility
    form_and_files_data.update(request.form)  # type: ignore[arg-type]
    return MultiDictProxy(form_and_files_data, schema)


class MarshmallowAdapter(SchemaAdapter):
    """Schema adapter for marshmallow schemas."""

    def __init__(
        self,
        schema: Schema | dict | type[Schema],
        schema_name: str | None = None,
        many: bool = False,
    ) -> None:
        """Initialize the marshmallow adapter.

        Arguments:
            schema: Marshmallow schema instance, dict, or schema class
            schema_name: Optional schema name for dict schemas
            many: Whether this schema represents a list/array of items
        """
        if isinstance(schema, dict):
            # Convert dict schema to marshmallow schema
            if schema_name is None:
                schema_name = 'GeneratedSchema'
            if schema == {}:
                self.schema = EmptySchema()
            else:
                self.schema = Schema.from_dict(schema, name=schema_name)()  # type: ignore
        elif isinstance(schema, type):
            # Check if it's a marshmallow schema class
            try:
                if issubclass(schema, Schema):
                    self.schema = schema()
                else:
                    # If not a marshmallow schema, keep as class (might be Pydantic)
                    self.schema = schema  # type: ignore[unreachable]
            except TypeError:
                # Not a class that can be checked with issubclass
                self.schema = schema
        else:
            self.schema = schema

        self.many = many

    @property
    def schema_type(self) -> str:
        return 'marshmallow'

    def validate_input(self, request: Request, location: str, **kwargs: t.Any) -> t.Any:
        """Validate input using marshmallow/webargs."""
        if location == 'files':
            # Handle file uploads with form data
            data = _get_files_and_form(request, self.schema)
            try:
                return self.schema.load(data)
            except MarshmallowValidationError as error:
                raise _ValidationError(
                    current_app.config['VALIDATION_ERROR_STATUS_CODE'],
                    current_app.config['VALIDATION_ERROR_DESCRIPTION'],
                    error.messages,
                ) from error

        # Use webargs for other locations
        return parser.load_location_data(schema=self.schema, req=request, location=location)

    def serialize_output(self, data: t.Any, many: bool = False) -> t.Any:
        """Serialize output using marshmallow."""
        if isinstance(self.schema, (EmptySchema, FileSchema)):
            return data

        # Use marshmallow's dump method which handles dump_default values
        return self.schema.dump(data, many=many)

    def get_openapi_schema(self, **kwargs: t.Any) -> dict[str, t.Any]:
        """Get OpenAPI schema from marshmallow schema."""
        if isinstance(self.schema, EmptySchema):
            return {}
        elif isinstance(self.schema, FileSchema):
            return {'type': self.schema.type, 'format': self.schema.format}

        from apispec.ext.marshmallow import MarshmallowPlugin

        # MarshmallowPlugin expects schema_name_resolver to be None or callable[[type[Schema]], str]
        # Since we don't need custom resolution here, pass None
        plugin = MarshmallowPlugin(schema_name_resolver=None)
        result: dict[str, t.Any] = plugin.schema_helper(self.schema, **kwargs)  # type: ignore[no-untyped-call]
        return result

    def get_schema_name(self) -> str:
        """Get schema name for OpenAPI documentation.

        *Version changed: 3.0.0*

        - Remove "Schema" suffix stripping from schema names.

        """
        schema = self.schema
        if isinstance(schema, type):  # pragma: no cover
            schema = schema()  # type: ignore

        name: str = schema.__class__.__name__
        if hasattr(schema, 'partial') and schema.partial:
            name += 'Update'
        return name



from dataclasses import dataclass

import openapi_spec_validator as osv
from flask import make_response
from flask.views import MethodView
from packaging.version import Version

from .conftest import APISPEC_VERSION
from .schemas import Foo
from .schemas import Query
from apiflask import Schema
from apiflask.fields import Field
from apiflask.fields import String


def test_output(app, client):
    @app.route('/foo')
    @app.output(Foo)
    def foo():
        return {'name': 'bar'}

    @app.route('/bar')
    @app.output(Foo, status_code=201)
    def bar():
        return {'name': 'foo'}

    @app.route('/baz')
    @app.input(Query, location='query')
    @app.output(Foo, status_code=201)
    def baz(query_data):
        if query_data['id'] == 1:
            return {'name': 'baz'}, 202
        elif query_data['id'] == 2:
            return {'name': 'baz'}, {'Location': '/baz'}
        elif query_data['id'] == 3:
            return {'name': 'baz'}, 202, {'Location': '/baz'}
        return ({'name': 'baz'},)

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json == {'id': 123, 'name': 'bar'}

    rv = client.get('/bar')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'foo'}

    rv = client.get('/baz')
    assert rv.status_code == 202
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert 'Location' not in rv.headers

    rv = client.get('/baz?id=2')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert rv.headers['Location'].endswith('/baz')

    rv = client.get('/baz?id=3')
    assert rv.status_code == 202
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert rv.headers['Location'].endswith('/baz')

    rv = client.get('/baz?id=4')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert 'Location' not in rv.headers


def test_output_with_methodview(app, client):
    @app.route('/')
    class FooAPI(MethodView):
        @app.output(Foo)
        def get(self):
            return {'name': 'bar'}

        @app.output(Foo, status_code=201)
        def post(self):
            return {'name': 'foo'}

        @app.input(Query, location='query')
        @app.output(Foo, status_code=201)
        def delete(self, query_data):
            if query_data['id'] == 1:
                return {'name': 'baz'}, 202
            elif query_data['id'] == 2:
                return {'name': 'baz'}, {'Location': '/baz'}
            elif query_data['id'] == 3:
                return {'name': 'baz'}, 202, {'Location': '/baz'}
            return ({'name': 'baz'},)

    rv = client.get('/')
    assert rv.status_code == 200
    assert rv.json == {'id': 123, 'name': 'bar'}

    rv = client.post('/')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'foo'}

    rv = client.delete('/')
    assert rv.status_code == 202
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert 'Location' not in rv.headers

    rv = client.delete('/?id=2')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert rv.headers['Location'].endswith('/baz')

    rv = client.delete('/?id=3')
    assert rv.status_code == 202
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert rv.headers['Location'].endswith('/baz')

    rv = client.delete('/?id=4')
    assert rv.status_code == 201
    assert rv.json == {'id': 123, 'name': 'baz'}
    assert 'Location' not in rv.headers


def test_output_with_dict_schema(app, client):
    dict_schema = {'name': String(dump_default='grey')}

    @app.get('/foo')
    @app.output(dict_schema, schema_name='MyName')
    def foo():
        return ''

    @app.get('/bar')
    @app.output(dict_schema, schema_name='MyName')
    def bar():
        return {'name': 'peter'}

    @app.get('/baz')
    @app.output(dict_schema)
    def baz():
        pass

    @app.get('/spam')
    @app.output(dict_schema)
    def spam():
        pass

    @app.get('/eggs')
    @app.output({})
    def eggs():
        pass

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json == {'name': 'grey'}

    rv = client.get('/bar')
    assert rv.status_code == 200
    assert rv.json == {'name': 'peter'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert (
        rv.json['paths']['/foo']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]['$ref']
        == '#/components/schemas/MyName'
    )

    # Check schema fields individually to avoid ordering issues
    schema = rv.json['components']['schemas']['MyName']
    assert schema['type'] == 'object'
    assert schema['properties'] == {'name': {'type': 'string'}}
    # In apispec >= 6.8.3, additionalProperties should be False (or omitted, treated as False)
    if APISPEC_VERSION >= Version('6.8.3'):
        assert schema.get('additionalProperties', False) is False
    else:
        assert 'additionalProperties' not in schema
    assert (
        rv.json['paths']['/bar']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]['$ref']
        == '#/components/schemas/MyName1'
    )
    # default schema name is "Generated"
    assert (
        rv.json['paths']['/baz']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]['$ref']
        == '#/components/schemas/GeneratedSchema'
    )
    assert (
        rv.json['paths']['/spam']['get']['responses']['200']['content']['application/json'][
            'schema'
        ]['$ref']
        == '#/components/schemas/GeneratedSchema1'
    )


def test_output_with_object_schema(app, client):
    class BaseResponse(Schema):
        data = Field()
        message = String(dump_default='Success')

    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponse

    class PetOut(Schema):
        name = String()

    @dataclass
    class Pet:
        name: str

    @dataclass
    class Response:
        data: Pet

    @app.get('/foo')
    @app.output(PetOut)
    def foo():
        pet = Pet('foo')
        return Response(data=pet)

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json['data'] == {'name': 'foo'}


def test_output_body_example(app, client):
    example = {'name': 'foo', 'id': 2}
    examples = {
        'example foo': {'summary': 'an example of foo', 'value': {'name': 'foo', 'id': 1}},
        'example bar': {'summary': 'an example of bar', 'value': {'name': 'bar', 'id': 2}},
    }

    @app.get('/foo')
    @app.output(Foo, example=example)
    def foo():
        pass

    @app.get('/bar')
    @app.output(Foo, examples=examples)
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert (
        rv.json['paths']['/foo']['get']['responses']['200']['content']['application/json'][
            'example'
        ]
        == example
    )
    assert (
        rv.json['paths']['/bar']['get']['responses']['200']['content']['application/json'][
            'examples'
        ]
        == examples
    )


def test_output_with_empty_dict_as_schema(app, client):
    @app.delete('/foo')
    @app.output({}, status_code=204)
    def delete_foo():
        return ''

    @app.route('/bar')
    class Bar(MethodView):
        @app.output({}, status_code=204)
        def delete(self):
            return ''

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'content' not in rv.json['paths']['/foo']['delete']['responses']['204']
    assert 'content' not in rv.json['paths']['/bar']['delete']['responses']['204']

    rv = client.delete('/foo')
    assert rv.status_code == 204
    rv = client.delete('/bar')
    assert rv.status_code == 204


def test_output_response_object_directly(app, client):
    @app.get('/foo')
    @app.output(Foo)
    def foo():
        return make_response({'message': 'hello'})

    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json['message'] == 'hello'


def test_response_links(app, client):
    links = {
        'foo': {'operationId': 'getFoo', 'parameters': {'id': 1}},
        'bar': {'operationId': 'getBar', 'parameters': {'id': 2}},
    }

    @app.get('/foo')
    @app.output(Foo, links=links)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['get']['responses']['200']['links'] == links


def test_response_links_ref(app, client):
    links = {'getFoo': {'$ref': '#/components/links/foo'}}

    @app.spec_processor
    def add_links(spec):
        spec['components']['links'] = {'foo': {'operationId': 'getFoo', 'parameters': {'id': 1}}}
        return spec

    @app.get('/foo')
    @app.output(Foo, links=links)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'getFoo' in rv.json['paths']['/foo']['get']['responses']['200']['links']


def test_response_content_type(app, client):
    @app.get('/foo')
    @app.output(Foo)  # default value is application/json
    def foo():
        pass

    @app.get('/bar')
    @app.output(Foo, content_type='image/png')
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert len(rv.json['paths']['/foo']['get']['responses']['200']['content']) == 1
    assert len(rv.json['paths']['/bar']['get']['responses']['200']['content']) == 1
    assert 'application/json' in rv.json['paths']['/foo']['get']['responses']['200']['content']
    assert 'image/png' in rv.json['paths']['/bar']['get']['responses']['200']['content']



from marshmallow import fields
from marshmallow import Schema
from marshmallow import validate

from apiflask.openapi_adapters import get_unique_schema_name
from apiflask.openapi_adapters import OpenAPIHelper


class SimpleSchema(Schema):
    """Simple test schema."""

    name = fields.String(required=True)
    age = fields.Integer()


class EnumSchema(Schema):
    """Schema with enum field."""

    status = fields.String(validate=validate.OneOf(['active', 'inactive']))


class NestedSchema(Schema):
    """Schema with nested field."""

    user = fields.Nested(SimpleSchema)
    tags = fields.List(fields.String())


def test_get_unique_schema_name():
    """Test get_unique_schema_name function."""
    from apispec import APISpec
    from apispec.ext.marshmallow import MarshmallowPlugin

    plugin = MarshmallowPlugin()
    spec = APISpec(title='Test', version='1.0.0', openapi_version='3.0.3', plugins=[plugin])

    # Register a schema
    spec.components.schema('TestSchema', schema=SimpleSchema())

    # Get unique name for conflicting schema
    unique_name = get_unique_schema_name(spec, 'TestSchema')
    assert unique_name == 'TestSchema1'

    # Register the unique name
    spec.components.schema(unique_name, schema=SimpleSchema())

    # Get another unique name
    unique_name2 = get_unique_schema_name(spec, 'TestSchema')
    assert unique_name2 == 'TestSchema2'


class TestOpenAPIHelper:
    """Test OpenAPIHelper class."""

    def test_get_marshmallow_plugin(self):
        """Test get_marshmallow_plugin method."""
        helper = OpenAPIHelper()
        plugin = helper.get_marshmallow_plugin()

        assert plugin is not None
        assert hasattr(plugin, 'converter')
        assert plugin.converter is not None

    def test_get_marshmallow_plugin_caching(self):
        """Test that marshmallow plugin is cached."""
        helper = OpenAPIHelper()
        plugin1 = helper.get_marshmallow_plugin()
        plugin2 = helper.get_marshmallow_plugin()

        assert plugin1 is plugin2

    def test_schema_to_spec_marshmallow(self):
        """Test schema_to_spec with marshmallow schema."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        spec = helper.schema_to_spec(schema)

        # schema_to_spec returns minimal spec (type only)
        # For full spec with properties, use schema_to_json_schema
        assert spec['type'] == 'object'

    def test_schema_to_spec_dict(self):
        """Test schema_to_spec with dict schema."""
        helper = OpenAPIHelper()
        schema = {'name': fields.String(required=True)}

        spec = helper.schema_to_spec(schema)

        # Dict schemas are converted to marshmallow schemas by the adapter
        assert spec['type'] == 'object'

    def test_schema_to_json_schema_marshmallow(self):
        """Test schema_to_json_schema with marshmallow schema."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        json_schema = helper.schema_to_json_schema(schema)

        assert json_schema['type'] == 'object'
        assert 'properties' in json_schema
        assert 'name' in json_schema['properties']
        assert json_schema['properties']['name']['type'] == 'string'
        assert 'age' in json_schema['properties']
        assert json_schema['properties']['age']['type'] == 'integer'

    def test_schema_to_parameters_query(self):
        """Test schema_to_parameters for query parameters."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        params = helper.schema_to_parameters(schema, location='query')

        assert len(params) == 2
        assert params[0]['name'] == 'name'
        assert params[0]['in'] == 'query'
        assert params[0]['required'] is True
        assert params[0]['schema']['type'] == 'string'

        assert params[1]['name'] == 'age'
        assert params[1]['in'] == 'query'
        assert params[1]['required'] is False

    def test_schema_to_parameters_headers(self):
        """Test schema_to_parameters for headers."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        params = helper.schema_to_parameters(schema, location='headers')

        assert len(params) == 2
        assert all(p['in'] == 'header' for p in params)

    def test_schema_to_parameters_path(self):
        """Test schema_to_parameters for path parameters."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        params = helper.schema_to_parameters(schema, location='view_args')

        assert len(params) == 2
        assert all(p['in'] == 'path' for p in params)

    def test_schema_to_parameters_with_enum(self):
        """Test schema_to_parameters handles enum validation."""
        helper = OpenAPIHelper()
        schema = EnumSchema()

        params = helper.schema_to_parameters(schema, location='query')

        assert len(params) == 1
        assert params[0]['schema']['type'] == 'string'
        assert 'enum' in params[0]['schema']
        assert params[0]['schema']['enum'] == ['active', 'inactive']

    def test_schema_to_parameters_location_mapping(self):
        """Test schema_to_parameters location mapping."""
        helper = OpenAPIHelper()
        schema = SimpleSchema()

        # Test various location mappings
        locations = {
            'query': 'query',
            'querystring': 'query',
            'headers': 'header',
            'view_args': 'path',
            'cookies': 'cookie',
        }

        for input_loc, expected_loc in locations.items():
            params = helper.schema_to_parameters(schema, location=input_loc)
            assert all(p['in'] == expected_loc for p in params)

    def test_get_schema_name_dict(self):
        """Test get_schema_name with dict schema."""
        helper = OpenAPIHelper()
        schema = {'name': fields.String()}

        name = helper.get_schema_name(schema)

        # Dict schemas get 'Generated' prefix
        assert name == 'GeneratedSchema'

    def test_schema_to_spec_unknown_type(self):
        """Test schema_to_spec with unknown schema type."""
        helper = OpenAPIHelper()

        # Pass something that's not a valid schema
        spec = helper.schema_to_spec('not_a_schema')

        # Should return fallback
        assert spec == {'type': 'object'}

    def test_schema_to_json_schema_unknown_type(self):
        """Test schema_to_json_schema with unknown schema type."""
        helper = OpenAPIHelper()

        # Pass something that's not a valid schema
        json_schema = helper.schema_to_json_schema('not_a_schema')

        # Should return fallback
        assert json_schema == {'type': 'object'}

    def test_schema_to_parameters_unknown_type(self):
        """Test schema_to_parameters with unknown schema type."""
        helper = OpenAPIHelper()

        # Pass something that's not a valid schema
        params = helper.schema_to_parameters('not_a_schema', location='query')

        # Should return empty list
        assert params == []



import pytest
from marshmallow import fields
from marshmallow import Schema
from marshmallow import ValidationError

from apiflask.schema_adapters import registry
from apiflask.schema_adapters.marshmallow import MarshmallowAdapter


class PetSchema(Schema):
    """Test pet schema."""

    name = fields.String(required=True)
    age = fields.Integer()
    species = fields.String(load_default='dog')


class UpdatePetSchema(Schema):
    """Partial schema for updates."""

    class Meta:
        partial = True

    name = fields.String()
    age = fields.Integer()


class TestAdapterRegistry:
    """Test AdapterRegistry class."""

    def test_registry_singleton(self):
        """Test that registry is a singleton."""
        from apiflask.schema_adapters import registry as registry1
        from apiflask.schema_adapters import registry as registry2

        assert registry1 is registry2

    def test_create_adapter_with_schema_instance(self):
        """Test creating adapter with schema instance."""
        schema = PetSchema()
        adapter = registry.create_adapter(schema)

        assert isinstance(adapter, MarshmallowAdapter)
        assert adapter.schema_type == 'marshmallow'

    def test_create_adapter_with_schema_class(self):
        """Test creating adapter with schema class."""
        adapter = registry.create_adapter(PetSchema)

        assert isinstance(adapter, MarshmallowAdapter)
        assert adapter.schema_type == 'marshmallow'

    def test_create_adapter_with_dict(self):
        """Test creating adapter with dict schema."""
        dict_schema = {'name': fields.String(required=True), 'age': fields.Integer()}
        adapter = registry.create_adapter(dict_schema)

        assert isinstance(adapter, MarshmallowAdapter)
        assert adapter.schema_type == 'marshmallow'

    def test_create_adapter_with_dict_custom_name(self):
        """Test creating adapter with dict schema and custom name."""
        dict_schema = {'name': fields.String(required=True)}
        adapter = registry.create_adapter(dict_schema, schema_name='CustomName')

        assert isinstance(adapter, MarshmallowAdapter)
        name = adapter.get_schema_name()
        assert name == 'CustomName'

    def test_create_adapter_handles_invalid_type(self):
        """Test that registry handles invalid types gracefully."""
        # Invalid types should raise an exception or return None
        try:
            adapter = registry.create_adapter('not_a_schema')
            # If it doesn't raise, it should handle gracefully
            assert adapter is not None
        except (TypeError, ValueError):
            # Expected to raise for invalid types
            pass


class TestMarshmallowAdapter:
    """Test MarshmallowAdapter class."""

    def test_init_with_schema_instance(self):
        """Test initializing with schema instance."""
        schema = PetSchema()
        adapter = MarshmallowAdapter(schema)

        assert adapter.schema is schema
        assert adapter.schema_type == 'marshmallow'

    def test_init_with_schema_class(self):
        """Test initializing with schema class."""
        adapter = MarshmallowAdapter(PetSchema)

        assert isinstance(adapter.schema, PetSchema)
        assert adapter.schema_type == 'marshmallow'

    def test_init_with_dict(self):
        """Test initializing with dict schema."""
        dict_schema = {'name': fields.String(required=True)}
        adapter = MarshmallowAdapter(dict_schema, schema_name='TestSchema')

        assert isinstance(adapter.schema, Schema)
        assert adapter.schema_type == 'marshmallow'

    def test_init_with_empty_dict(self):
        """Test initializing with empty dict."""
        from apiflask.schemas import EmptySchema

        adapter = MarshmallowAdapter({})

        assert isinstance(adapter.schema, EmptySchema)

    def test_get_schema_name_partial(self):
        """Test get_schema_name with partial schema."""
        schema = PetSchema(partial=True)
        adapter = MarshmallowAdapter(schema)

        name = adapter.get_schema_name()

        # Partial schemas should append 'Update'
        assert name == 'PetSchemaUpdate'

    # Note: get_openapi_schema is tested via openapi_helper in integration tests
    # Direct adapter usage is internal and may vary

    def test_get_openapi_schema_empty(self):
        """Test get_openapi_schema with empty schema."""
        adapter = MarshmallowAdapter({})

        openapi_schema = adapter.get_openapi_schema()

        assert openapi_schema == {}

    def test_get_openapi_schema_file_schema(self):
        """Test get_openapi_schema with FileSchema."""
        from apiflask.schemas import FileSchema

        adapter = MarshmallowAdapter(FileSchema())

        openapi_schema = adapter.get_openapi_schema()

        assert openapi_schema['type'] == 'string'
        assert openapi_schema['format'] == 'binary'

    def test_validate_input_json(self):
        """Test validate_input for json location."""
        from flask import Flask
        from werkzeug.test import EnvironBuilder
        from flask import request

        app = Flask(__name__)
        app.config['VALIDATION_ERROR_STATUS_CODE'] = 422
        app.config['VALIDATION_ERROR_DESCRIPTION'] = 'Validation error'
        adapter = MarshmallowAdapter(PetSchema())

        with app.test_request_context():
            builder = EnvironBuilder(
                method='POST', json={'name': 'Fluffy', 'age': 3}, content_type='application/json'
            )
            env = builder.get_environ()

            with app.request_context(env):
                data = adapter.validate_input(request, 'json')

                assert data['name'] == 'Fluffy'
                assert data['age'] == 3
                # load_default behavior depends on marshmallow configuration

    def test_validate_input_query(self):
        """Test validate_input for query location."""
        from flask import Flask

        app = Flask(__name__)
        app.config['VALIDATION_ERROR_STATUS_CODE'] = 422
        app.config['VALIDATION_ERROR_DESCRIPTION'] = 'Validation error'
        adapter = MarshmallowAdapter(PetSchema())

        with app.test_request_context('/?name=Buddy&age=5'):
            from flask import request

            data = adapter.validate_input(request, 'query')

            assert data['name'] == 'Buddy'
            # Query params come as strings, may need conversion
            assert 'age' in data

    def test_serialize_output(self):
        """Test serialize_output method."""
        adapter = MarshmallowAdapter(PetSchema())

        data = {'name': 'Max', 'age': 7, 'extra': 'ignored'}
        output = adapter.serialize_output(data)

        assert output['name'] == 'Max'
        assert output['age'] == 7
        # Extra fields should be excluded
        assert 'extra' not in output

    def test_serialize_output_many(self):
        """Test serialize_output with many=True."""
        adapter = MarshmallowAdapter(PetSchema())

        data = [{'name': 'Max', 'age': 7}, {'name': 'Bella', 'age': 3}]
        output = adapter.serialize_output(data, many=True)

        assert len(output) == 2
        assert output[0]['name'] == 'Max'
        assert output[1]['name'] == 'Bella'

    def test_serialize_output_empty_schema(self):
        """Test serialize_output with EmptySchema."""
        adapter = MarshmallowAdapter({})

        data = {'any': 'data'}
        output = adapter.serialize_output(data)

        # EmptySchema should pass through data unchanged
        assert output == data

    def test_serialize_output_file_schema(self):
        """Test serialize_output with FileSchema."""
        from apiflask.schemas import FileSchema

        adapter = MarshmallowAdapter(FileSchema())

        data = b'file content'
        output = adapter.serialize_output(data)

        # FileSchema should pass through data unchanged
        assert output == data


class TestFlaskParser:
    """Test FlaskParser class."""

    def test_flask_parser_singleton(self):
        """Test that parser is a singleton."""
        from apiflask.schema_adapters.marshmallow import parser as parser1
        from apiflask.schema_adapters.marshmallow import parser as parser2

        assert parser1 is parser2

    def test_handle_error(self):
        """Test handle_error method."""
        from flask import Flask
        from apiflask.exceptions import _ValidationError
        from apiflask.schema_adapters.marshmallow import parser

        app = Flask(__name__)
        app.config['VALIDATION_ERROR_STATUS_CODE'] = 400
        app.config['VALIDATION_ERROR_DESCRIPTION'] = 'Bad Request'

        with app.app_context():
            error = ValidationError({'field': ['error message']})
            schema = PetSchema()

            with pytest.raises(_ValidationError) as exc_info:
                parser.handle_error(error, None, schema, error_status_code=400, error_headers={})

            assert exc_info.value.status_code == 400
            assert exc_info.value.message == 'Bad Request'
            assert exc_info.value.detail == {'field': ['error message']}

    def test_load_location_data(self):
        """Test load_location_data method."""
        from flask import Flask
        from apiflask.schema_adapters.marshmallow import parser

        app = Flask(__name__)
        schema = PetSchema()

        with app.test_request_context('/?name=Buddy&age=5'):
            from flask import request

            data = parser.load_location_data(schema=schema, req=request, location='query')

            assert data['name'] == 'Buddy'
            # Query params may be strings or converted
            assert 'age' in data

    def test_location_loader_files(self):
        """Test files location loader."""
        from flask import Flask
        from io import BytesIO

        app = Flask(__name__)

        class FileUploadSchema(Schema):
            file = fields.Field()
            description = fields.String()

        schema = FileUploadSchema()

        with app.test_request_context(
            method='POST',
            data={
                'file': (BytesIO(b'file content'), 'test.txt'),
                'description': 'A test file',
            },
            content_type='multipart/form-data',
        ):
            from flask import request
            from apiflask.schema_adapters.marshmallow import parser

            data = parser.load_location_data(schema=schema, req=request, location='files')

            assert 'file' in data
            assert data['description'] == 'A test file'

    def test_location_loader_form_and_files(self):
        """Test form_and_files location loader."""
        from flask import Flask
        from io import BytesIO

        app = Flask(__name__)

        class FormWithFileSchema(Schema):
            file = fields.Field()
            name = fields.String()

        schema = FormWithFileSchema()

        with app.test_request_context(
            method='POST',
            data={'file': (BytesIO(b'content'), 'file.txt'), 'name': 'test'},
            content_type='multipart/form-data',
        ):
            from flask import request
            from apiflask.schema_adapters.marshmallow import parser

            data = parser.load_location_data(schema=schema, req=request, location='form_and_files')

            assert 'file' in data
            assert data['name'] == 'test'
